#ifndef _XOR_ACC_H
#define _XOR_ACC_H

#include <iostream>
#include <iomanip>
#include <sstream>

#include <systemc>
#include "tlm_utils/simple_target_socket.h"
#include "tlm_utils/simple_initiator_socket.h"

// #include "hybrid_tlm.h"

// Caution: The verilator related headers must include after SystemC headers.
//          Otherwise it will finds no SYSTEMC_VERSION and assume you
//          are using oldest SystemC API.
#include "Vxor_acc.h"

using namespace sc_core;
using namespace tlm;

typedef union word {
    int sint;
    unsigned int uint;
    unsigned char uc[4];
} word;

SC_MODULE(Xor_acc)
{
    tlm_utils::simple_target_socket<Xor_acc> targ_skt;

    SC_CTOR(Xor_acc):
        targ_skt("xor_acc_target_socket"), m_clk_period(sc_time(1, SC_NS))
    {
        // reset phase
        m_xor_acc.reset = 1;
        // make a rising edge to clock
        step();
        // end of reset phase
        m_xor_acc.reset = 0;
        // register the TLM transport function
        targ_skt.register_b_transport(this, &Xor_acc::b_transport);
        SC_THREAD(Xor_acc_process);
    }

    void step() {
        // make a rising edge to clock
        m_xor_acc.clock = 0;
        m_xor_acc.eval();
        m_xor_acc.clock = 1;
        m_xor_acc.eval();
    }

    void Xor_acc_process()
    {
        while (true)
        {
            unsigned int cycle = 0;
            word input, key;
            // read data
            input.uint = num_fifo.read();
            m_xor_acc.io_data_0 = input.uc[0];
            m_xor_acc.io_data_1 = input.uc[1];
            m_xor_acc.io_data_2 = input.uc[2];
            m_xor_acc.io_data_3 = input.uc[3];
            input.uint = num_fifo.read();
            m_xor_acc.io_data_4 = input.uc[0];
            m_xor_acc.io_data_5 = input.uc[1];
            m_xor_acc.io_data_6 = input.uc[2];
            m_xor_acc.io_data_7 = input.uc[3];
            // read key
            input.uint = num_fifo.read();
            m_xor_acc.io_key_0 = input.uc[0];
            m_xor_acc.io_key_1 = input.uc[1];
            m_xor_acc.io_key_2 = input.uc[2];
            m_xor_acc.io_key_3 = input.uc[3];
            input.uint = num_fifo.read();
            m_xor_acc.io_key_4 = input.uc[0];
            m_xor_acc.io_key_5 = input.uc[1];
            m_xor_acc.io_key_6 = input.uc[2];
            m_xor_acc.io_key_7 = input.uc[3];
            step();
            cycle++;
            wait(cycle * m_clk_period);

            word ans;
            // write ans
            ans.uc[0] = m_xor_acc.io_xor_0;
            ans.uc[1] = m_xor_acc.io_xor_1;
            ans.uc[2] = m_xor_acc.io_xor_2;
            ans.uc[3] = m_xor_acc.io_xor_3;
            result_fifo.write(ans.uint);
            ans.uc[0] = m_xor_acc.io_xor_4;
            ans.uc[1] = m_xor_acc.io_xor_5;
            ans.uc[2] = m_xor_acc.io_xor_6;
            ans.uc[3] = m_xor_acc.io_xor_7;
            result_fifo.write(ans.uint);
        }
    }

    // Parse the payload and read from the FIFO or write to the FIFO
    void b_transport( tlm_generic_payload & trans, sc_time & delay )
    {
        tlm_command cmd = trans.get_command();
        // unsigned *       ptr = reinterpret_cast<unsigned *>(trans.get_data_ptr());
        // unsigned int     len = trans.get_data_length();
        unsigned char *data_ptr = trans.get_data_ptr();

        word in_package;
        word out_package;
        switch (cmd) {
        case TLM_WRITE_COMMAND:
            // num_fifo.write(*ptr);
            in_package.uc[0] = data_ptr[0];
            in_package.uc[1] = data_ptr[1];
            in_package.uc[2] = data_ptr[2];
            in_package.uc[3] = data_ptr[3];
            num_fifo.write(in_package.uint);
            wait(1 * m_clk_period);
            break;

        case TLM_READ_COMMAND:
            out_package.uint = result_fifo.read();
            data_ptr[0] = out_package.uc[0];
            data_ptr[1] = out_package.uc[1];
            data_ptr[2] = out_package.uc[2];
            data_ptr[3] = out_package.uc[3];
            break;

        default:
            break;
        }

        trans.set_response_status( TLM_OK_RESPONSE );
    }

private:
    Vxor_acc m_xor_acc;
    sc_fifo<unsigned int> num_fifo;
    sc_fifo<unsigned int> result_fifo;
    const sc_time m_clk_period;
};


#endif

