#include "my_sobel.h"

#define MASK_N 2
#define MASK_X 3
#define MASK_Y 3
#define WHITE  255
#define BLACK  0

unsigned char *image_s = NULL;     // source image array
unsigned char *image_t = NULL;     // target image array

unsigned int   width, height;      // image width, image height
unsigned int   rgb_raw_data_offset;// RGB raw data offset
unsigned char  bit_per_pixel;      // bit per pixel
unsigned short byte_per_pixel;     // byte per pixel
int left_quarter_to_sobel = 4;

void My_sobel::read_bmp() {
    FILE *fp_s = NULL;                 // source file handler
    fp_s = fopen("lena.bmp", "rb");
    if (fp_s == NULL) {
        printf("fopen fp_s error\n");
        return;
    }

    // move offset to 10 to find rgb raw data offset
    fseek(fp_s, 10, SEEK_SET);
    fread(&rgb_raw_data_offset, sizeof(unsigned int), 1, fp_s);

    // move offset to 18 to get width & height;
    fseek(fp_s, 18, SEEK_SET);
    fread(&width,  sizeof(unsigned int), 1, fp_s);
    fread(&height, sizeof(unsigned int), 1, fp_s);

    // get bit per pixel
    fseek(fp_s, 28, SEEK_SET);
    fread(&bit_per_pixel, sizeof(unsigned short), 1, fp_s);
    byte_per_pixel = bit_per_pixel / 8;

    // move offset to rgb_raw_data_offset to get RGB raw data
    fseek(fp_s, rgb_raw_data_offset, SEEK_SET);

    image_s = (unsigned char *)malloc((size_t)width * height * byte_per_pixel);
    if (image_s == NULL) {
        printf("malloc images_s error\n");
        return;
    }

    image_t = (unsigned char *)malloc((size_t)width * height * byte_per_pixel);
    if (image_t == NULL) {
        printf("malloc image_t error\n");
        return;
    }

    fread(image_s, sizeof(unsigned char), (size_t)(long)width * height * byte_per_pixel, fp_s);
    fclose(fp_s);

    printf("readDone_event.notify()\n");
    wait(1, SC_NS);
    readDone_event.notify();
}

// convert RGB to gray level int
int color_to_int(int r, int g, int b) {
    return (r + g + b) / 3;
}

void My_sobel::sobel() {
    unsigned int  x, y, i, v, u;             // for loop counter
    unsigned char R, G, B;         // color of R, G, B
    double val[MASK_N] = {0.0};
    int adjustX, adjustY, xBound, yBound;
    double total;
    double threshold = 90.0;
    // sobel mask
    int mask[MASK_N][MASK_X][MASK_Y] = {
        {   { -1, -2, -1},
            {0 , 0, 0},
            {1 , 2, 1}
        },

        {   { -1, 0, 1},
            { -2, 0, 2},
            { -1, 0, 1}
        }
    };

    unsigned int x_offset;
    unsigned int y_offset;

    wait(readDone_event);

    while (true) {
        if (left_quarter_to_sobel < 4)
            wait(again_event);
        printf("sobel in while\n");

        if (left_quarter_to_sobel == 4) {
            x_offset = width / 2;
            y_offset = 0;
        }
        else if (left_quarter_to_sobel == 3) {
            x_offset = 0;
            y_offset = height / 2;
        }
        else if (left_quarter_to_sobel == 2) {
            x_offset = width / 2;
            y_offset = height / 2;
        }
        else if (left_quarter_to_sobel == 1) {
            x_offset = 0;
            y_offset = 0;
        }
        for (y = 0 + y_offset; y != height / 2 + y_offset; ++y) {
            for (x = 0 + x_offset; x != width / 2 + x_offset; ++x) {
                for (i = 0; i != MASK_N; ++i) {
                    adjustX = (MASK_X % 2) ? 1 : 0;
                    adjustY = (MASK_Y % 2) ? 1 : 0;
                    xBound = MASK_X / 2;
                    yBound = MASK_Y / 2;

                    val[i] = 0.0;
                    for (v = -yBound; v != yBound + adjustY; ++v) {
                        for (u = -xBound; u != xBound + adjustX; ++u) {
                            if (x + u >= 0 + x_offset && x + u < width / 2 + x_offset && y + v >= 0 + y_offset && y + v < height / 2 + y_offset) {
                                R = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 2);
                                G = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 1);
                                B = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 0);

                                val[i] +=    color_to_int(R, G, B) * mask[i][u + xBound][v + yBound];
                            }
                        }
                    }
                }

                total = 0.0;
                for (i = 0; i != MASK_N; ++i) {
                    total += val[i] * val[i];
                }

                total = sqrt(total);

                if (total - threshold >= 0) {
                    // black
                    *(image_t + byte_per_pixel * (width * y + x) + 2) = BLACK;
                    *(image_t + byte_per_pixel * (width * y + x) + 1) = BLACK;
                    *(image_t + byte_per_pixel * (width * y + x) + 0) = BLACK;
                } else {
                    // white
                    *(image_t + byte_per_pixel * (width * y + x) + 2) = WHITE;
                    *(image_t + byte_per_pixel * (width * y + x) + 1) = WHITE;
                    *(image_t + byte_per_pixel * (width * y + x) + 0) = WHITE;
                }
            }
        }
        left_quarter_to_sobel -= 1;
        wait(1, SC_NS);
        sobelDone_event.notify();

    }
}

void My_sobel::write_bmp() {
    unsigned int file_size; // file size
    // bitmap header
    unsigned char header[54] = {
        0x42,        // identity : B
        0x4d,        // identity : M
        0, 0, 0, 0,  // file size
        0, 0,        // reserved1
        0, 0,        // reserved2
        54, 0, 0, 0, // RGB data offset
        40, 0, 0, 0, // struct BITMAPINFOHEADER size
        0, 0, 0, 0,  // bmp width
        0, 0, 0, 0,  // bmp height
        1, 0,        // planes
        24, 0,       // bit per pixel
        0, 0, 0, 0,  // compression
        0, 0, 0, 0,  // data size
        0, 0, 0, 0,  // h resolution
        0, 0, 0, 0,  // v resolution
        0, 0, 0, 0,  // used colors
        0, 0, 0, 0   // important colors
    };

    while (true) {
        wait(sobelDone_event);
        printf("write\n");
        FILE *fp_t = NULL;                 // target file handler
        fp_t = fopen("lena_sobel.bmp", "wb");
        if (fp_t == NULL) {
            printf("fopen fname_t error\n");
            return;
        }

        // file size
        file_size = width * height * byte_per_pixel + rgb_raw_data_offset;
        header[2] = (unsigned char)(file_size & 0x000000ff);
        header[3] = (file_size >> 8)  & 0x000000ff;
        header[4] = (file_size >> 16) & 0x000000ff;
        header[5] = (file_size >> 24) & 0x000000ff;

        // width
        header[18] = width & 0x000000ff;
        header[19] = (width >> 8)  & 0x000000ff;
        header[20] = (width >> 16) & 0x000000ff;
        header[21] = (width >> 24) & 0x000000ff;

        // height
        header[22] = height & 0x000000ff;
        header[23] = (height >> 8)  & 0x000000ff;
        header[24] = (height >> 16) & 0x000000ff;
        header[25] = (height >> 24) & 0x000000ff;

        // bit per pixel
        header[28] = bit_per_pixel;

        // write header
        fwrite(header, sizeof(unsigned char), rgb_raw_data_offset, fp_t);

        // write image
        fwrite(image_t, sizeof(unsigned char), (size_t)(long)width * height * byte_per_pixel, fp_t);
        fclose(fp_t);

        if (left_quarter_to_sobel > 0) {
            wait(1, SC_NS);
            again_event.notify();
        }

    }
}

