// convert RGB to gray level int
#include "top.h"
#include "sobel.h"

#define MASK_N 2
#define MASK_X 3
#define MASK_Y 3
#define WHITE  255
#define BLACK  0

int color_to_int(int r, int g, int b) {
    return (r + g + b) / 3;
}

void Sobel::sobel() {
    unsigned int  x, y, i, v, u;             // for loop counter
    unsigned char R, G, B;         // color of R, G, B
    double val[MASK_N] = {0.0};
    int adjustX, adjustY, xBound, yBound;
    double total;
    double threshold = 90.0;
    // sobel mask
    int mask[MASK_N][MASK_X][MASK_Y] = {
        {   { -1, -2, -1},
            {0 , 0, 0},
            {1 , 2, 1}
        },

        {   { -1, 0, 1},
            { -2, 0, 2},
            { -1, 0, 1}
        }
    };

    unsigned int x_offset;
    unsigned int y_offset;

    wait(*i_readDone_event);

    width = i_width.read();
    height = i_height.read();
    rgb_raw_data_offset = i_rgb_raw_data_offset.read();
    bit_per_pixel = i_bit_per_pixel.read();
    byte_per_pixel = i_byte_per_pixel.read();
    image_s = i_image_s.read();
    image_t = i_image_t.read();
    left_quarter_to_sobel = 4;

    while (true) {
        if (left_quarter_to_sobel < 4)
            wait(*i_again_event);
        printf("sobel in while\n");

        if (left_quarter_to_sobel == 4) {
            x_offset = width / 2;
            y_offset = 0;
        }
        else if (left_quarter_to_sobel == 3) {
            x_offset = 0;
            y_offset = height / 2;
        }
        else if (left_quarter_to_sobel == 2) {
            x_offset = width / 2;
            y_offset = height / 2;
        }
        else if (left_quarter_to_sobel == 1) {
            x_offset = 0;
            y_offset = 0;
        }
        for (y = 0 + y_offset; y != height / 2 + y_offset; ++y) {
            for (x = 0 + x_offset; x != width / 2 + x_offset; ++x) {
                for (i = 0; i != MASK_N; ++i) {
                    adjustX = (MASK_X % 2) ? 1 : 0;
                    adjustY = (MASK_Y % 2) ? 1 : 0;
                    xBound = MASK_X / 2;
                    yBound = MASK_Y / 2;

                    val[i] = 0.0;
                    for (v = -yBound; v != yBound + adjustY; ++v) {
                        for (u = -xBound; u != xBound + adjustX; ++u) {
                            if (x + u >= 0 + x_offset && x + u < width / 2 + x_offset && y + v >= 0 + y_offset && y + v < height / 2 + y_offset) {
                                R = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 2);
                                G = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 1);
                                B = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 0);

                                val[i] +=    color_to_int(R, G, B) * mask[i][u + xBound][v + yBound];
                            }
                        }
                    }
                }

                total = 0.0;
                for (i = 0; i != MASK_N; ++i) {
                    total += val[i] * val[i];
                }

                total = sqrt(total);

                if (total - threshold >= 0) {
                    // black
                    *(image_t + byte_per_pixel * (width * y + x) + 2) = BLACK;
                    *(image_t + byte_per_pixel * (width * y + x) + 1) = BLACK;
                    *(image_t + byte_per_pixel * (width * y + x) + 0) = BLACK;
                } else {
                    // white
                    *(image_t + byte_per_pixel * (width * y + x) + 2) = WHITE;
                    *(image_t + byte_per_pixel * (width * y + x) + 1) = WHITE;
                    *(image_t + byte_per_pixel * (width * y + x) + 0) = WHITE;
                }
            }
        }
        left_quarter_to_sobel -= 1;
        o_width.write(width);
        o_height.write(height);
        o_rgb_raw_data_offset.write(rgb_raw_data_offset);
        o_bit_per_pixel.write(bit_per_pixel);
        o_byte_per_pixel.write(byte_per_pixel);
        o_image_t.write(image_t);
        o_left_quarter_to_sobel.write(left_quarter_to_sobel);
        wait(1, SC_NS);
        o_sobelDone_event->notify();

    }
}