#ifndef _TOP_H_
#define _TOP_H_

#include <systemc>

#include "read_bmp.h"
#include "sobel.h"
#include "write_bmp.h"

using namespace sc_core;

class Read_bmp;
class Sobel;
class Write_bmp;


SC_MODULE(Top) {
	SC_CTOR(Top) {
		read_ptr = new Read_bmp("read_bmp");
		sobel_ptr = new Sobel("sobel");
		write_ptr = new Write_bmp("write_bmp");

		read_ptr->o_width(r_s_width);
		read_ptr->o_height(r_s_height);
		read_ptr->o_rgb_raw_data_offset(r_s_rgb_raw_data_offset);
		read_ptr->o_bit_per_pixel(r_s_bit_per_pixel);
		read_ptr->o_byte_per_pixel(r_s_byte_per_pixel);
		read_ptr->o_image_s(r_s_image_s);
		read_ptr->o_image_t(r_s_image_t);

		sobel_ptr->i_width(r_s_width);
		sobel_ptr->i_height(r_s_height);
		sobel_ptr->i_rgb_raw_data_offset(r_s_rgb_raw_data_offset);
		sobel_ptr->i_bit_per_pixel(r_s_bit_per_pixel);
		sobel_ptr->i_byte_per_pixel(r_s_byte_per_pixel);
		sobel_ptr->i_image_s(r_s_image_s);
		sobel_ptr->i_image_t(r_s_image_t);

		sobel_ptr->o_width(s_w_width);
		sobel_ptr->o_height(s_w_height);
		sobel_ptr->o_rgb_raw_data_offset(s_w_rgb_raw_data_offset);
		sobel_ptr->o_bit_per_pixel(s_w_bit_per_pixel);
		sobel_ptr->o_byte_per_pixel(s_w_byte_per_pixel);
		sobel_ptr->o_image_t(s_w_image_t);
		sobel_ptr->o_left_quarter_to_sobel(s_w_left_quarter_to_sobel);

		write_ptr->i_width(s_w_width);
		write_ptr->i_height(s_w_height);
		write_ptr->i_rgb_raw_data_offset(s_w_rgb_raw_data_offset);
		write_ptr->i_bit_per_pixel(s_w_bit_per_pixel);
		write_ptr->i_byte_per_pixel(s_w_byte_per_pixel);
		write_ptr->i_image_t(s_w_image_t);
		write_ptr->i_left_quarter_to_sobel(s_w_left_quarter_to_sobel);

		read_ptr->o_readDone_event = &readDone_event;
		sobel_ptr->i_readDone_event = &readDone_event;
		sobel_ptr->o_sobelDone_event = &sobelDone_event;
		write_ptr->i_sobelDone_event = &sobelDone_event;
		write_ptr->o_again_event = &again_event;
		sobel_ptr->i_again_event = &again_event;
	}

	sc_fifo<unsigned int>   r_s_width;
	sc_fifo<unsigned int>   r_s_height;
	sc_fifo<unsigned int>   r_s_rgb_raw_data_offset;
	sc_fifo<unsigned char>  r_s_bit_per_pixel;
	sc_fifo<unsigned short> r_s_byte_per_pixel;
	sc_fifo<uint8_t*>       r_s_image_s;
	sc_fifo<uint8_t*>       r_s_image_t;

	sc_fifo<unsigned int>   s_w_width;
	sc_fifo<unsigned int>   s_w_height;
	sc_fifo<unsigned int>   s_w_rgb_raw_data_offset;
	sc_fifo<unsigned char>  s_w_bit_per_pixel;
	sc_fifo<unsigned short> s_w_byte_per_pixel;
	sc_fifo<uint8_t*>       s_w_image_t;
	sc_fifo<unsigned int>   s_w_left_quarter_to_sobel;

	sc_event readDone_event, sobelDone_event, again_event;

	Read_bmp *read_ptr;
	Sobel *sobel_ptr;
	Write_bmp *write_ptr;
};

#endif
