#include "Sobel.h"
#include "Initiator.h"


#define MASK_N 2
#define MASK_X 3
#define MASK_Y 3
#define WHITE  255
#define BLACK  0

Sobel::Sobel( sc_module_name n ): sc_module( n )
{
    // register the function after_sobel_transport() to t_skt
    SC_THREAD( sobel );
}

int color_to_int(int r, int g, int b) {
    return (r + g + b) / 3;
}

void Sobel::sobel() {
    unsigned int  x, y, i, v, u;   // for loop counter
    unsigned char R, G, B;         // color of R, G, B
    double val[MASK_N] = {0.0};
    int adjustX, adjustY, xBound, yBound;
    double total;
    double threshold = 90.0;
    // sobel mask
    int mask[MASK_N][MASK_X][MASK_Y] = {
        {   { -1, -2, -1},
            {0 , 0, 0},
            {1 , 2, 1}
        },

        {   { -1, 0, 1},
            { -2, 0, 2},
            { -1, 0, 1}
        }
    };

    unsigned int x_offset;
    unsigned int y_offset;

    // read data from fifo
    width = i_uint_width_height_rgb.read();
    height = i_uint_width_height_rgb.read();
    rgb_raw_data_offset = i_uint_width_height_rgb.read();
    bit_per_pixel = i_uchar_bit_byte.read();
    byte_per_pixel = i_uchar_bit_byte.read();
    image_s = i_uchar_ptr_images_imaget.read();
    image_t = i_uchar_ptr_images_imaget.read();

    // left parts to sobel
    int left = 4;

    while ( left > 0 ) {
        if (left == 1) {
            x_offset = width / 2;
            y_offset = 0;
        }
        else if (left == 2) {
            x_offset = 0;
            y_offset = height / 2;
        }
        else if (left == 4) {
            x_offset = width / 2;
            y_offset = height / 2;
        }
        else if (left == 3) {
            x_offset = 0;
            y_offset = 0;
        }
        for (y = 0 + y_offset; y != height / 2 + y_offset; ++y) {
            for (x = 0 + x_offset; x != width / 2 + x_offset; ++x) {
                for (i = 0; i != MASK_N; ++i) {
                    adjustX = (MASK_X % 2) ? 1 : 0;
                    adjustY = (MASK_Y % 2) ? 1 : 0;
                    xBound = MASK_X / 2;
                    yBound = MASK_Y / 2;

                    val[i] = 0.0;
                    for (v = -yBound; v != yBound + adjustY; ++v) {
                        for (u = -xBound; u != xBound + adjustX; ++u) {
                            if (x + u >= 0 + x_offset && x + u < width / 2 + x_offset && y + v >= 0 + y_offset && y + v < height / 2 + y_offset) {
                                R = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 2);
                                G = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 1);
                                B = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 0);

                                val[i] +=    color_to_int(R, G, B) * mask[i][u + xBound][v + yBound];
                            }
                        }
                    }
                }

                total = 0.0;
                for (i = 0; i != MASK_N; ++i) {
                    total += val[i] * val[i];
                }

                total = sqrt(total);

                if (total - threshold >= 0) {
                    // black
                    *(image_t + byte_per_pixel * (width * y + x) + 2) = BLACK;
                    *(image_t + byte_per_pixel * (width * y + x) + 1) = BLACK;
                    *(image_t + byte_per_pixel * (width * y + x) + 0) = BLACK;
                } else {
                    // white
                    *(image_t + byte_per_pixel * (width * y + x) + 2) = WHITE;
                    *(image_t + byte_per_pixel * (width * y + x) + 1) = WHITE;
                    *(image_t + byte_per_pixel * (width * y + x) + 0) = WHITE;
                }
            }
        }

		// std::cout << "---------sobel done" << std::endl;
                
		wait(1, SC_NS);
		// std::cout << name() <<" is done at "<< sc_time_stamp() << std::endl;
		
		o_uint_width_height_rgb.write(width);
        o_uint_width_height_rgb.write(height);
        o_uint_width_height_rgb.write(rgb_raw_data_offset);
        o_uchar_bit_byte.write(bit_per_pixel);
        o_uchar_bit_byte.write(byte_per_pixel);
        o_uchar_ptr_images_imaget.write(image_s);
        o_uchar_ptr_images_imaget.write(image_t);

        left -= 1;
    }
}
