#include "SobelTLM.h"
#include "Sobel.h"
#include "Initiator.h"
#include "sobel_tlm.h"
#include <iomanip>

SobelTLM::SobelTLM( sc_module_name n ): sc_module( n ), t_skt_before_sobel( "t_skt_before_sobel" ), t_skt_after_sobel( "t_skt_after_sobel" ), base_offset( 0 )
{
    sobel_ptr = new Sobel("sobel");

    sobel_ptr->i_uint_width_height_rgb( fifo_rs_uint_width_height_rgb );
    sobel_ptr->i_uchar_bit_byte( fifo_rs_uchar_bit_byte );
    sobel_ptr->i_uchar_ptr_images_imaget( fifo_rs_uchar_ptr_images_imaget );

    sobel_ptr->o_uint_width_height_rgb( fifo_sw_uint_width_height_rgb );
    sobel_ptr->o_uchar_bit_byte( fifo_sw_uchar_bit_byte );
    sobel_ptr->o_uchar_ptr_images_imaget( fifo_sw_uchar_ptr_images_imaget );

    // register the function before_sobel_transport() to t_skt
    t_skt_before_sobel.register_b_transport( this, &SobelTLM::before_sobel_transport );
    t_skt_after_sobel.register_b_transport( this, &SobelTLM::after_sobel_transport );
}

void SobelTLM::before_sobel_transport( tlm::tlm_generic_payload &payload, sc_core::sc_time &delay )
{
    sc_dt::uint64 addr = payload.get_address();
    addr = addr - base_offset;
    unsigned char *mask_ptr = payload.get_byte_enable_ptr();
    unsigned char *data_ptr = payload.get_data_ptr();

    word input_buffer;
    switch ( payload.get_command() )
    {
    case tlm::TLM_READ_COMMAND:
        /* write only */
        break;
    case tlm::TLM_WRITE_COMMAND:
        switch ( addr )
        {
        case SOBEL_INPUT_WIDTH_OFFSET: {
            input_buffer.uc[0] = data_ptr[0];
            input_buffer.uc[1] = data_ptr[1];
            input_buffer.uc[2] = data_ptr[2];
            input_buffer.uc[3] = data_ptr[3];
            fifo_rs_uint_width_height_rgb.write( input_buffer.uint );
            break;
        }
        case SOBEL_INPUT_BIT_OFFSET: {
            if (mask_ptr[0] == 0xff) { fifo_rs_uchar_bit_byte.write(data_ptr[0]); }
            if (mask_ptr[1] == 0xff) { fifo_rs_uchar_bit_byte.write(data_ptr[1]); }
            break;
        }
        case SOBEL_INPUT_IMAGES_OFFSET: {
            image_info* info;
            info = ( image_info* )malloc( sizeof( image_info ) );
            memcpy( info, data_ptr, sizeof(image_info) );
            fifo_rs_uchar_ptr_images_imaget.write(info->image_s);
            fifo_rs_uchar_ptr_images_imaget.write(info->image_t);
            break;
        }
        default:
            std::cerr << "Error! JPEGEncoderTLM::r_g_b_transport: address 0x"
                      << std::setfill('0') << std::setw(8) << std::hex << addr << std::dec
                      << " is not valid" << std::endl;
            break;
        }
        break;
    case tlm::TLM_IGNORE_COMMAND:
        payload.set_response_status( tlm::TLM_GENERIC_ERROR_RESPONSE );
        return;
    default:
        payload.set_response_status( tlm::TLM_GENERIC_ERROR_RESPONSE );
        return;
    }
    payload.set_response_status( tlm::TLM_OK_RESPONSE );  // Always OK
}

void SobelTLM::after_sobel_transport( tlm::tlm_generic_payload &payload, sc_core::sc_time &delay )
{
    // std::cout << "after_sobel_transport start" << std::endl;
    sc_dt::uint64 addr = payload.get_address();
    addr = addr - base_offset;
    unsigned char *mask_ptr = payload.get_byte_enable_ptr();
    unsigned char *data_ptr = payload.get_data_ptr();

    image_info* info;
    info = ( image_info* )malloc( sizeof( image_info ) );
    word output_buffer;

    switch ( payload.get_command() )
    {
    case tlm::TLM_READ_COMMAND:
        switch ( addr )
        {
        case SOBEL_OUTPUT_WIDTH_OFFSET:
            output_buffer.uint = fifo_sw_uint_width_height_rgb.read();
            break;
        case SOBEL_OUTPUT_BIT_OFFSET:
            output_buffer.uc[0] = fifo_sw_uchar_bit_byte.read();
            output_buffer.uc[1] = fifo_sw_uchar_bit_byte.read();
            break;
        case SOBEL_OUTPUT_IMAGES_OFFSET:
            info->image_s = fifo_sw_uchar_ptr_images_imaget.read();
            info->image_t = fifo_sw_uchar_ptr_images_imaget.read();
            memcpy( data_ptr, info, sizeof(image_info) );
            break;
        default:
            std::cerr << "Error! JPEGEncoderTLM::code_bits_transport: address 0x"
                      << std::setfill('0') << std::setw(8) << std::hex << addr << std::dec
                      << " is not valid" << std::endl;
            break;
        }
        data_ptr[0] = output_buffer.uc[0];
        data_ptr[1] = output_buffer.uc[1];
        data_ptr[2] = output_buffer.uc[2];
        data_ptr[3] = output_buffer.uc[3];
        break;
    case tlm::TLM_WRITE_COMMAND:
        /* read only */
        break;
    case tlm::TLM_IGNORE_COMMAND:
        payload.set_response_status( tlm::TLM_GENERIC_ERROR_RESPONSE );
        return;
    default:
        payload.set_response_status( tlm::TLM_GENERIC_ERROR_RESPONSE );
        return;
    }
    payload.set_response_status( tlm::TLM_OK_RESPONSE );  // Always OK
}
