#ifndef _SOBEL_H
#define _SOBEL_H

#include <iostream>
#include <iomanip>
#include <sstream>

#include <systemc>
#include "tlm_utils/simple_target_socket.h"
#include "tlm_utils/simple_initiator_socket.h"

#include "sobel_tlm.h"

// Caution: The verilator related headers must include after SystemC headers.
//          Otherwise it will finds no SYSTEMC_VERSION and assume you
//          are using oldest SystemC API.
#include "Vsobel.h"

using namespace sc_core;
using namespace tlm;

// const char* MSGID = "Sobel-TLM-Wrapper";


// struct ReturnValue {
//   unsigned int iVal1;
//   unsigned int iVal2;
// };

// typedef union word {
//   int sint;
//   unsigned int uint;
//   unsigned char uc[4];
// } word;


// void print_result_time_format(unsigned int a, unsigned int b, unsigned int c, unsigned int d,
//                               const char* comment_of_number) {
//   PRINT_TIME_FORMAT(comment_of_number << ": " << a << " " << b << " " << c << " " << d)
// }

// The GCD function used for verification
// unsigned int SoftGCD(unsigned int u, unsigned int v) {
//     while ( v != 0) {
//         unsigned int r = u % v;
//         u = v;
//         v = r;
//     }
//     return u;
// }

// ReturnValue SoftSobel(unsigned char m0, unsigned char m1, unsigned char m2, unsigned char m3, unsigned char m5, unsigned char m6, unsigned char m7, unsigned char m8) {
//   ReturnValue SWrv;
//   unsigned int row_0 = m0 + 2 * m1 + m2;
//   unsigned int row_2 = m6 + 2 * m7 + m8;
//   unsigned int col_0 = m0 + 2 * m3 + m6;
//   unsigned int col_2 = m2 + 2 * m5 + m8;
//   if (row_0 > row_2) {
//     SWrv.iVal1 = row_0 - row_2;
//   }
//   else {
//     SWrv.iVal1 = row_2 - row_0;
//   }
//   if (col_0 > col_2) {
//     SWrv.iVal2 = col_0 - col_2;
//   }
//   else {
//     SWrv.iVal2 = col_2 - col_0;
//   }
//   return SWrv;
// }

SC_MODULE(Sobel)
{
    tlm_utils::simple_target_socket<Sobel> targ_skt;

    SC_CTOR(Sobel):
        targ_skt("sobel_target_socket"), m_clk_period(sc_time(1, SC_NS))
    {
        // reset phase
        m_sobel.reset = 1;
        // make a rising edge to clock
        step();
        // end of reset phase
        m_sobel.reset = 0;
        // register the TLM transport function
        targ_skt.register_b_transport(this, &Sobel::b_transport);
        SC_THREAD(Sobel_process);
    }

    void step() {
        // make a rising edge to clock
        m_sobel.clock = 0;
        m_sobel.eval();
        m_sobel.clock = 1;
        m_sobel.eval();
    }

    void Sobel_process()
    {
        while (true)
        {
            word package;
            package.uint = num_fifo.read();
            m_sobel.io_m0 = package.uc[3];
            m_sobel.io_m1 = package.uc[2];
            m_sobel.io_m2 = package.uc[1];
            m_sobel.io_m3 = package.uc[0];

            package.uint = num_fifo.read();
            m_sobel.io_m5 = package.uc[3];
            m_sobel.io_m6 = package.uc[2];
            m_sobel.io_m7 = package.uc[1];
            m_sobel.io_m8 = package.uc[0];

            unsigned int cycle = 0;
            step();
            cycle++;
            wait(cycle * m_clk_period);
            unsigned int result1 = m_sobel.io_out_X;
            unsigned int result2 = m_sobel.io_out_Y;
            result_fifo.write(result1);
            result_fifo.write(result2);
            // result_fifo.write(m_sobel.io_out_X);
            // result_fifo.write(m_sobel.io_out_Y);
        }
    }

    // Parse the payload and read from the FIFO or write to the FIFO
    void b_transport( tlm_generic_payload & trans, sc_time & delay )
    {
        tlm_command cmd = trans.get_command();
        // unsigned *       ptr = reinterpret_cast<unsigned *>(trans.get_data_ptr());
        // unsigned int     len = trans.get_data_length();
        unsigned char *data_ptr = trans.get_data_ptr();

        word in_package;
        word out_package;
        switch (cmd) {
        case TLM_WRITE_COMMAND:
            // num_fifo.write(*ptr);
            in_package.uc[3] = data_ptr[3];
            in_package.uc[2] = data_ptr[2];
            in_package.uc[1] = data_ptr[1];
            in_package.uc[0] = data_ptr[0];
            num_fifo.write(in_package.uint);
            wait(1 * m_clk_period);
            break;

        case TLM_READ_COMMAND:
            out_package.uint = result_fifo.read();
            data_ptr[3] = out_package.uc[3];
            data_ptr[2] = out_package.uc[2];
            data_ptr[1] = out_package.uc[1];
            data_ptr[0] = out_package.uc[0];
            break;

        default:
            break;
        }

        trans.set_response_status( TLM_OK_RESPONSE );
    }

private:
    Vsobel m_sobel;
    sc_fifo<unsigned int> num_fifo;
    sc_fifo<unsigned int> result_fifo;
    const sc_time m_clk_period;
};

#endif

// SC_MODULE( Tester )
// {
//   tlm_utils::simple_initiator_socket<Tester> init_skt;

//   SC_CTOR( Tester ):
//     init_skt( "tester_init_socket" )
//   {
//     SC_THREAD(test_process);
//   }

//   void test_process() {
//     // unsigned int a = 64;
//     // unsigned int b = 48;
//     unsigned char m0, m1, m2, m3, m5, m6, m7, m8;
//     unsigned seed;
//     srand((unsigned)time(NULL));
//     m0 = rand() % 255;
//     m1 = rand() % 255;
//     m2 = rand() % 255;
//     m3 = rand() % 255;
//     m5 = rand() % 255;
//     m6 = rand() % 255;
//     m7 = rand() % 255;
//     m8 = rand() % 255;
//     send_numbers(m0, m1, m2, m3);
//     send_numbers(m5, m6, m7, m8);
//     // send_numbers(a, b);
//     auto result_X = recv_result();
//     auto result_Y = recv_result();

//     ReturnValue golden = SoftSobel(m0, m1, m2, m3, m5, m6, m7, m8);
//     // auto golden = SoftSobel(m0, m1, m2, m3, m5, m6, m7, m8);

//     if ((result_Y == golden.iVal1) && (result_X == golden.iVal2)) {
//       std::cout << "Test Passed.\n";
//     } else {
//       std::cout << "Test Failed!\n";
//     }
//   }

// // Prepare the write payload and send it to GcdTlm
// void send_numbers(const unsigned char a, const unsigned char b, const unsigned char c, const unsigned char d) {
//   tlm::tlm_generic_payload* trans = new tlm::tlm_generic_payload;
//   sc_time delay = sc_time( 0, SC_NS );
//   // unsigned int numbers = (a << 16) | (b);

//   word package;
//   package.uc[3] = a;
//   package.uc[2] = b;
//   package.uc[1] = c;
//   package.uc[0] = d;
//   unsigned int numbers = package.uint;

//   trans->set_command( TLM_WRITE_COMMAND );
//   trans->set_data_ptr(reinterpret_cast<unsigned char*>(&numbers));
//   trans->set_data_length(4);
//   // print_result_time_format(a, b, c, d, "Tester sending numbers");
//   init_skt->b_transport( *trans, delay );
// }

// // Prepare the read payload and send it to GcdTlm
// unsigned int recv_result() {
//   tlm::tlm_generic_payload* trans = new tlm::tlm_generic_payload;
//   sc_time delay = sc_time( 0, SC_NS );
//   unsigned int result;
//   trans->set_command( TLM_READ_COMMAND );
//   trans->set_data_ptr(reinterpret_cast<unsigned char*>(&result));
//   trans->set_data_length(4);
//   // PRINT_TIME_FORMAT("Tester reading result.")
//   init_skt->b_transport( *trans, delay );
//   // PRINT_TIME_FORMAT("Tester read result: " << result)
//   return result;
// }



// int sc_main(int argc, char *argv[])
// {
//   Sobel sobel("Sobel");
//   Tester tester("Tester");
//   tester.init_skt.bind(sobel.targ_skt);
//   sc_start();
//   return 0;
// }
