#include "SobelTLM.h"
#include "Sobel.h"
#include "Initiator.h"
#include "sobel_tlm.h"
#include <iomanip>

#define MASK_N 2
#define MASK_X 3
#define MASK_Y 3
#define WHITE  255
#define BLACK  0

int color_to_int(int r, int g, int b) {
    return (r + g + b) / 3;
}

SobelTLM::SobelTLM( sc_module_name n ): sc_module( n ), t_skt_sobel( "t_skt_sobel" ), initiator( "initiator" ), base_offset( 0 )
{
    SC_THREAD( sobel );
    t_skt_sobel.register_b_transport( this, &SobelTLM::sobel_transport );
    // t_skt_after_sobel.register_b_transport( this, &SobelTLM::after_sobel_transport );
}

void SobelTLM::sobel() {
    unsigned int  x, y, i, v, u;   // for loop counter
    unsigned char R, G, B;         // color of R, G, B
    // double val[MASK_N] = {0.0};
    // int adjustX, adjustY, xBound, yBound;
    double total;
    double threshold = 90.0;

    unsigned int x_offset;
    unsigned int y_offset;

    // read data from fifo
    width = fifo_rs_uint_width_height_rgb.read();
    height = fifo_rs_uint_width_height_rgb.read();
    rgb_raw_data_offset = fifo_rs_uint_width_height_rgb.read();
    bit_per_pixel = fifo_rs_uchar_bit_byte.read();
    byte_per_pixel = fifo_rs_uchar_bit_byte.read();
    image_s = fifo_rs_uchar_ptr_images_imaget.read();
    image_t = fifo_rs_uchar_ptr_images_imaget.read();

    // left parts to sobel
    int left = 4;

    while ( left > 0 ) {
        if (left == 1) {
            x_offset = width / 2;
            y_offset = 0;
        }
        else if (left == 2) {
            x_offset = 0;
            y_offset = height / 2;
        }
        else if (left == 4) {
            x_offset = width / 2;
            y_offset = height / 2;
        }
        else if (left == 3) {
            x_offset = 0;
            y_offset = 0;
        }
        for (y = 0 + y_offset; y != height / 2 + y_offset; ++y) {
            for (x = 0 + x_offset; x != width / 2 + x_offset; ++x) {
                // for (i = 0; i != MASK_N; ++i) {
                //     adjustX = (MASK_X % 2) ? 1 : 0;
                //     adjustY = (MASK_Y % 2) ? 1 : 0;
                //     xBound = MASK_X / 2;
                //     yBound = MASK_Y / 2;
                double val[2] = {0.0};
                unsigned char tmp[9] = { 0 };
                int cnt = 0;
                // val[i] = 0.0;
                for (v = -1; v != 1 + 1; ++v) {
                    for (u = -1; u != 1 + 1; ++u) {
                        if (x + u >= 0 + x_offset && x + u < width / 2 + x_offset && y + v >= 0 + y_offset && y + v < height / 2 + y_offset) {
                            R = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 2);
                            G = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 1);
                            B = *(image_s + byte_per_pixel * (width * (y + v) + (x + u)) + 0);

                            tmp[cnt] = color_to_int(R, G, B);
                            // val[i] +=    color_to_int(R, G, B) * mask[i][u + 1][v + 1];
                        }
                        cnt++;
                    }
                }
                send_numbers(tmp[0], tmp[1], tmp[2], tmp[3]);
                send_numbers(tmp[5], tmp[6], tmp[7], tmp[8]);
                val[0] = recv_result();
                val[1] = recv_result();
                // }

                total = 0.0;
                for (i = 0; i != MASK_N; ++i) {
                    total += val[i] * val[i];
                }

                total = sqrt(total);

                if (total - threshold >= 0) {
                    // black
                    *(image_t + byte_per_pixel * (width * y + x) + 2) = BLACK;
                    *(image_t + byte_per_pixel * (width * y + x) + 1) = BLACK;
                    *(image_t + byte_per_pixel * (width * y + x) + 0) = BLACK;
                } else {
                    // white
                    *(image_t + byte_per_pixel * (width * y + x) + 2) = WHITE;
                    *(image_t + byte_per_pixel * (width * y + x) + 1) = WHITE;
                    *(image_t + byte_per_pixel * (width * y + x) + 0) = WHITE;
                }
            }
        }

        wait(1, SC_NS);

        fifo_sw_uint_width_height_rgb.write(width);
        fifo_sw_uint_width_height_rgb.write(height);
        fifo_sw_uint_width_height_rgb.write(rgb_raw_data_offset);
        fifo_sw_uchar_bit_byte.write(bit_per_pixel);
        fifo_sw_uchar_bit_byte.write(byte_per_pixel);
        fifo_sw_uchar_ptr_images_imaget.write(image_s);
        fifo_sw_uchar_ptr_images_imaget.write(image_t);

        left -= 1;
        std::cout<< name() <<" is done at "<< sc_time_stamp() << std::endl;
    }
}

void SobelTLM::sobel_transport( tlm::tlm_generic_payload &payload, sc_core::sc_time &delay )
{
    sc_dt::uint64 addr = payload.get_address();
    addr = addr - base_offset;
    // unsigned char *mask_ptr = payload.get_byte_enable_ptr();
    unsigned char *data_ptr = payload.get_data_ptr();

    image_info* info;
    info = ( image_info* )malloc( sizeof( image_info ) );
    word output_buffer;

    word input_buffer;
    switch ( payload.get_command() )
    {
    case tlm::TLM_READ_COMMAND:
        switch ( addr )
        {
        case SOBEL_OUTPUT_WIDTH_OFFSET:
            output_buffer.uint = fifo_sw_uint_width_height_rgb.read();
            break;
        case SOBEL_OUTPUT_BIT_OFFSET:
            output_buffer.uc[0] = fifo_sw_uchar_bit_byte.read();
            output_buffer.uc[1] = fifo_sw_uchar_bit_byte.read();
            break;
        case SOBEL_OUTPUT_IMAGES_OFFSET:
            info->image_s = fifo_sw_uchar_ptr_images_imaget.read();
            info->image_t = fifo_sw_uchar_ptr_images_imaget.read();
            memcpy( data_ptr, info, sizeof(image_info) );
            break;
        default:
            std::cerr << "Error! JPEGEncoderTLM::code_bits_transport: address 0x"
                      << std::setfill('0') << std::setw(8) << std::hex << addr << std::dec
                      << " is not valid" << std::endl;
            break;
        }
        data_ptr[0] = output_buffer.uc[0];
        data_ptr[1] = output_buffer.uc[1];
        data_ptr[2] = output_buffer.uc[2];
        data_ptr[3] = output_buffer.uc[3];
        break;
    case tlm::TLM_WRITE_COMMAND:
        switch ( addr )
        {
        case SOBEL_INPUT_WIDTH_OFFSET: {
            input_buffer.uc[0] = data_ptr[0];
            input_buffer.uc[1] = data_ptr[1];
            input_buffer.uc[2] = data_ptr[2];
            input_buffer.uc[3] = data_ptr[3];
            fifo_rs_uint_width_height_rgb.write( input_buffer.uint );
            break;
        }
        case SOBEL_INPUT_BIT_OFFSET: {
            fifo_rs_uchar_bit_byte.write(data_ptr[0]);
            fifo_rs_uchar_bit_byte.write(data_ptr[1]);
            break;
        }
        case SOBEL_INPUT_IMAGES_OFFSET: {
            image_info* info;
            info = ( image_info* )malloc( sizeof( image_info ) );
            memcpy( info, data_ptr, sizeof(image_info) );
            fifo_rs_uchar_ptr_images_imaget.write(info->image_s);
            fifo_rs_uchar_ptr_images_imaget.write(info->image_t);
            break;
        }
        default:
            std::cerr << "Error! JPEGEncoderTLM::r_g_b_transport: address 0x"
                      << std::setfill('0') << std::setw(8) << std::hex << addr << std::dec
                      << " is not valid" << std::endl;
            break;
        }
        break;
    case tlm::TLM_IGNORE_COMMAND:
        payload.set_response_status( tlm::TLM_GENERIC_ERROR_RESPONSE );
        return;
    default:
        payload.set_response_status( tlm::TLM_GENERIC_ERROR_RESPONSE );
        return;
    }
    payload.set_response_status( tlm::TLM_OK_RESPONSE );  // Always OK
}

// Prepare the write payload and send it to GcdTlm
void SobelTLM::send_numbers(const unsigned char a, const unsigned char b, const unsigned char c, const unsigned char d) {
    // tlm::tlm_generic_payload* trans = new tlm::tlm_generic_payload;
    // sc_time delay = sc_time( 0, SC_NS );
    // unsigned int numbers = (a << 16) | (b);
    unsigned char mask_int[4] = {0xff, 0xff, 0xff, 0xff};
    word package;
    package.uc[3] = a;
    package.uc[2] = b;
    package.uc[1] = c;
    package.uc[0] = d;
    // unsigned int numbers = package.uint;

    // trans->set_command( TLM_WRITE_COMMAND );
    // trans->set_data_ptr(reinterpret_cast<unsigned char*>(&numbers));
    // trans->set_data_length(4);
    // // print_result_time_format(a, b, c, d, "Tester sending numbers");
    // init_skt->b_transport( *trans, delay );
    initiator.write_to_socket( HWSOBEL_BASE + HWSOBEL_INPUT_OFFSET, mask_int, package.uc, 4);
}

// Prepare the read payload and send it to GcdTlm
unsigned int SobelTLM::recv_result() {
    // tlm::tlm_generic_payload* trans = new tlm::tlm_generic_payload;
    // sc_time delay = sc_time( 0, SC_NS );
    // trans->set_command( TLM_READ_COMMAND );
    // trans->set_data_ptr(reinterpret_cast<unsigned char*>(&result));
    // trans->set_data_length(4);
    // // PRINT_TIME_FORMAT("Tester reading result.")
    // init_skt->b_transport( *trans, delay );
    // PRINT_TIME_FORMAT("Tester read result: " << result)
    unsigned char mask_int[4] = {0xff, 0xff, 0xff, 0xff};
    word package;
    initiator.read_from_socket( HWSOBEL_BASE + HWSOBEL_OUTPUT_OFFSET, mask_int, package.uc, 4);
    unsigned int result = package.uint;
    return result;
}
