#include "WriteTLM.h"
#include "sobel_tlm.h"

WriteTLM::WriteTLM( sc_module_name n ): sc_module( n ), initiator( "initiator" ), base_offset( 0 ),
    sobel_output_width_addr( SOBEL_BASE + SOBEL_OUTPUT_WIDTH_OFFSET ),
    sobel_output_height_addr( SOBEL_BASE + SOBEL_OUTPUT_HEIGHT_OFFSET ),
    sobel_output_rgb_addr( SOBEL_BASE + SOBEL_OUTPUT_RGB_OFFSET ),
    sobel_output_bit_addr( SOBEL_BASE + SOBEL_OUTPUT_BIT_OFFSET ),
    sobel_output_byte_addr( SOBEL_BASE + SOBEL_OUTPUT_BYTE_OFFSET ),
    sobel_output_images_addr( SOBEL_BASE + SOBEL_OUTPUT_IMAGES_OFFSET ),
    sobel_output_imaget_addr( SOBEL_BASE + SOBEL_OUTPUT_IMAGET_OFFSET )
{
    // register the function after_sobel_transport() to t_skt
    SC_THREAD( write_bmp );
}

void WriteTLM::write_bmp() {
    unsigned int file_size; // file size
    // bitmap header
    unsigned char header[54] = {
        0x42,        // identity : B
        0x4d,        // identity : M
        0, 0, 0, 0,  // file size
        0, 0,        // reserved1
        0, 0,        // reserved2
        54, 0, 0, 0, // RGB data offset
        40, 0, 0, 0, // struct BITMAPINFOHEADER size
        0, 0, 0, 0,  // bmp width
        0, 0, 0, 0,  // bmp height
        1, 0,        // planes
        24, 0,       // bit per pixel
        0, 0, 0, 0,  // compression
        0, 0, 0, 0,  // data size
        0, 0, 0, 0,  // h resolution
        0, 0, 0, 0,  // v resolution
        0, 0, 0, 0,  // used colors
        0, 0, 0, 0   // important colors
    };

    int left = 4;
    while ( left > 0)
    {
        FILE *fp_t = NULL;                 // target file handler
        fp_t = fopen("lena_sobel.bmp", "wb");
        if (fp_t == NULL) {
            printf("fopen fname_t error\n");
            return;
        }

        word tmp;
        unsigned char mask_tmp[4] = {0xff, 0xff, 0xff, 0xff};

        initiator.read_from_socket( sobel_output_width_addr, mask_tmp, tmp.uc, 4 );
        width = tmp.uint;
        initiator.read_from_socket( sobel_output_width_addr, mask_tmp, tmp.uc, 4 );
        height = tmp.uint;
        initiator.read_from_socket( sobel_output_width_addr, mask_tmp, tmp.uc, 4 );
        rgb_raw_data_offset = tmp.uint;
        initiator.read_from_socket( sobel_output_bit_addr, mask_tmp, tmp.uc, 4 );
        bit_per_pixel = tmp.uc[0];
        byte_per_pixel = tmp.uc[1];

        image_info info;
        mask = (unsigned char *) malloc(sizeof(image_info));
        for ( int i = 0; i < int(sizeof(image_info) / sizeof(unsigned char)); i++ )
            mask[i] = 0xff;
        initiator.read_from_socket( sobel_output_images_addr, mask, (unsigned char*)(&info), sizeof(image_info) / sizeof(unsigned char));

        image_t = info.image_t;

        // file size
        file_size = width * height * byte_per_pixel + rgb_raw_data_offset;
        header[2] = (unsigned char)(file_size & 0x000000ff);
        header[3] = (file_size >> 8)  & 0x000000ff;
        header[4] = (file_size >> 16) & 0x000000ff;
        header[5] = (file_size >> 24) & 0x000000ff;

        // width
        header[18] = width & 0x000000ff;
        header[19] = (width >> 8)  & 0x000000ff;
        header[20] = (width >> 16) & 0x000000ff;
        header[21] = (width >> 24) & 0x000000ff;

        // height
        header[22] = height & 0x000000ff;
        header[23] = (height >> 8)  & 0x000000ff;
        header[24] = (height >> 16) & 0x000000ff;
        header[25] = (height >> 24) & 0x000000ff;

        // bit per pixel
        header[28] = bit_per_pixel;

        // write header
        fwrite(header, sizeof(unsigned char), rgb_raw_data_offset, fp_t);


        // image_t = fifo_image_t.read();

        // write image
        fwrite(image_t, sizeof(unsigned char), (size_t)(long)width * height * byte_per_pixel, fp_t);
        fclose(fp_t);
        left -= 1;
        wait(1, SC_NS);
        std::cout << name() <<" is done at "<< sc_time_stamp() << std::endl;
    }

}
