# package name : SystemC

# SystemC_FOUND
# SystemC_ROOT_DIR
# SystemC_LIBRARIES
# SystemC_INCLUDE_DIRS

set(possible_systemc_dir_name systemc-2.3.1 systemc-2.3 systemc)

if(NOT DEFINED SystemC_FIND_QUIETLY)
	set(SystemC_FIND_QUIETLY false)
endif()

if(NOT COMMAND append_list_each)
	message(FATAL_ERROR "please include utils.cmake before find_package()")
endif(NOT COMMAND append_list_each)

execute_process(COMMAND uname -m
		OUTPUT_VARIABLE MACH
		OUTPUT_STRIP_TRAILING_WHITESPACE)

# If on 64bit platform, choose Linux64 first.
if(${MACH} MATCHES 64)
	set(SYSTEMC_R_PATH "lib-linux64")
else()
	set(SYSTEMC_R_PATH "lib-linux")
endif()

# find the SystemC library
foreach(dir_name ${possible_systemc_dir_name})
	append_list_each(systemc_path_list CMAKE_PREFIX_PATH /${dir_name})
	list(APPEND POSSIBLE_SystemC_PATHS ${systemc_path_list})
endforeach(dir_name)


foreach(SEARCH_PATH ${POSSIBLE_SystemC_PATHS})
	find_library(SystemC_LIBRARIES systemc
			PATH_SUFFIXES ${SYSTEMC_R_PATH}
			PATHS ${SEARCH_PATH} NO_DEFAULT_PATH)

	if(${SystemC_LIBRARIES} STREQUAL "SystemC_LIBRARIES-NOTFOUND")
		set(SystemC_FOUND false)
	else()
		set(SystemC_FOUND true)
	endif()

	if(${SystemC_FOUND})
		get_filename_component(SystemC_MACH_DIR ${SystemC_LIBRARIES} PATH)
		set(temp_include_dir "${SystemC_MACH_DIR}/../include")
		try_run(run_result_var compile_result_var
					${CMAKE_BINARY_DIR}/test_dir 
					"${SCRIPT_DIR}/TestCode/testSystemC.cpp"
					COMPILE_DEFINITIONS -I${temp_include_dir}
					LINK_LIBRARIES ${SystemC_LIBRARIES}
					RUN_OUTPUT_VARIABLE output_var)

		if( ${compile_result_var} AND (${run_result_var} EQUAL 0) )
			break()
		else()
			set(SystemC_FOUND false)
		endif()
	endif()

	unset(SystemC_LIBRARIES CACHE)
endforeach(SEARCH_PATH)


set(SystemC_FOUND ${SystemC_FOUND} CACHE BOOL "is found SystemC")

if(${SystemC_FOUND})

get_filename_component(SystemC_MACH_DIR ${SystemC_LIBRARIES} PATH)
get_filename_component(SystemC_ROOT_DIR ${SystemC_MACH_DIR} PATH CACHE)

if(NOT ${SystemC_FIND_QUIETLY})
	message(STATUS "SystemC ${SystemC_VERSION} found : ${SystemC_ROOT_DIR}")
endif(NOT ${SystemC_FIND_QUIETLY})

set(SystemC_INCLUDE_DIRS "${SystemC_ROOT_DIR}/include")
set(SystemC_INCLUDE_DIRS "${SystemC_INCLUDE_DIRS}"
					"${SystemC_INCLUDE_DIRS}/tlm_utils"
					CACHE PATH "the included directory")

else(${SystemC_FOUND})
	if(${SystemC_FIND_REQUIRED})
		message(FATAL_ERROR "finds no SystemC library")
	endif(${SystemC_FIND_REQUIRED})
endif(${SystemC_FOUND})

