#include <iostream>
#include "Vsobel.h"

// sets up a macro to help printing
#define PRINT_IO(io,type) {std::cout << #io << ": " << (type) dut.io << '\n';}

struct ReturnValue {
    unsigned int iVal1;
    unsigned int iVal2;
};

// IO printers
void printIO(Vsobel& dut, unsigned int cycle) {
    std::cout << "------ IO at time "  << cycle << " ----------\n";

    // Here are IO registers
    // PRINT_IO(io_a, unsigned)
    // PRINT_IO(io_b, unsigned)
    // PRINT_IO(io_e, bool)
    // PRINT_IO(io_v, bool)
    // PRINT_IO(io_z, unsigned)


    PRINT_IO(io_m0, unsigned);
    PRINT_IO(io_m1, unsigned);
    PRINT_IO(io_m2, unsigned);
    PRINT_IO(io_m3, unsigned);
    PRINT_IO(io_m5, unsigned);
    PRINT_IO(io_m6, unsigned);
    PRINT_IO(io_m7, unsigned);
    PRINT_IO(io_m8, unsigned);
    PRINT_IO(io_out_X, unsigned);
    PRINT_IO(io_out_Y, unsigned);


    // Here are internal registers
    // PRINT_IO(GCD__DOT__x, unsigned)
    // PRINT_IO(GCD__DOT__y, unsigned)
}

void step(Vsobel& dut) {
    // make a rising edge to clock
    dut.clock = 0;
    dut.eval();
    dut.clock = 1;
    dut.eval();
}

// Feed inputs to the hardware IO
// parameters:
//     cycle: pass by reference, used to return the simulated cycle count
//     a, b : pass by value, inputs of the chisel circuit
ReturnValue HardSobel(unsigned int& cycle, unsigned int m0, unsigned int m1, unsigned int m2, unsigned int m3, unsigned int m5, unsigned int m6, unsigned int m7, unsigned int m8) {
    // create the Chisel module
    Vsobel dut;
    // reset phase
    dut.reset = 1;
    // make a rising edge to clock
    step(dut);
    // end of reset phase
    dut.reset = 0;
    // write inputs and wait for one cycle to make sure the
    // value is writen to internal registers,
    dut.io_m0 = m0;
    dut.io_m1 = m1;
    dut.io_m2 = m2;
    dut.io_m3 = m3;
    dut.io_m5 = m5;
    dut.io_m6 = m6;
    dut.io_m7 = m7;
    dut.io_m8 = m8;
    // dut.io_e = 1;
    // make a rising edge to clock
    step(dut);
    // increment the time
    cycle++;
    // turn off the enable so that the internal registers would
    // not be overwritten by IO registers
    // dut.io_e = 0;

    printIO(dut, cycle);
    // Check whether the valid signal rise
    // while(! dut.io_v) {
    //     // make a rising edge to clock
    //     step(dut);
    //     // increment the time
    //     cycle++;
    //     printIO(dut, cycle);
    // }

    // return the greatest divided digit
    ReturnValue HWrv;
    HWrv.iVal1 = dut.io_out_X;
    HWrv.iVal2 = dut.io_out_Y;
    return HWrv;
}

// The GCD function used for verification
ReturnValue SoftSobel(unsigned int m0, unsigned int m1, unsigned int m2, unsigned int m3, unsigned int m5, unsigned int m6, unsigned int m7, unsigned int m8) {
    ReturnValue SWrv;
    // SWrv.iVal1 = abs((m6 + 2 * m7 + m8) - (m0 + 2 * m1 + m2));
    // SWrv.iVal2 = abs((m2 + 2 * m5 + m8) - (m0 + 2 * m3 + m6));
    unsigned int row_0 = m0 + 2 * m1 + m2;
    unsigned int row_2 = m6 + 2 * m7 + m8;
    unsigned int col_0 = m0 + 2 * m3 + m6;
    unsigned int col_2 = m2 + 2 * m5 + m8;
    if (row_0 > row_2) {
        SWrv.iVal1 = row_0 - row_2;
    }
    else {
        SWrv.iVal1 = row_2 - row_0;
    }
    if (col_0 > col_2) {
        SWrv.iVal2 = col_0 - col_2;
    }
    else {
        SWrv.iVal2 = col_2 - col_0;
    }

    return SWrv;
}



int main() {
    unsigned int cycle = 0;
    unsigned int m0, m1, m2, m3, m5, m6, m7, m8;
    unsigned seed;
    srand((unsigned)time(NULL));
    m0 = rand() % 255;
    m1 = rand() % 255;
    m2 = rand() % 255;
    m3 = rand() % 255;
    m5 = rand() % 255;
    m6 = rand() % 255;
    m7 = rand() % 255;
    m8 = rand() % 255;

    ReturnValue HWrv = HardSobel(cycle, m0, m1, m2, m3, m5, m6, m7, m8);
    ReturnValue SWrv = SoftSobel(m0, m1, m2, m3, m5, m6, m7, m8);

    // unsigned int HardwareResult = HardSobel(cycle, m0, m1, m2, m3, m5, m6, m7, m8);

    std::cout << "solve " << "(" << m0 << ", " << m1 << ", " << m2 << ", " << m3 << ", " << m5 << ", " << m6 << ", " << m7 << ", " << m8 << "): " << HWrv.iVal1 << ", " << HWrv.iVal2
              << ", expected:  " << SWrv.iVal1 << ", " << SWrv.iVal2
              << ", cycles: " << cycle << '\n';

    return 0;
}
