/*
 *
 * Copyright (c) 2005-2012 Imperas Software Ltd., www.imperas.com
 *
 * The contents of this file are provided under the Software License
 * Agreement that you accepted before downloading this file.
 *
 * This source forms part of the Software and can be used for educational,
 * training, and demonstration purposes but cannot be used for derivative
 * works except in cases where the derivative works require OVP technology
 * to run.
 *
 * For open source models released under licenses that you can use for
 * derivative works, please visit www.OVPworld.org or www.imperas.com
 * for the location of the open source models.
 *
 */


////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                W R I T T E N   B Y   I M P E R A S   I G E N               //
//                                                                            //
//                          Wed Apr 28 08:42:21 2010                          //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#include "tlm.h"
#include "ovpworld.org/modelSupport/tlmPlatform/1.0/tlm2.0/tlmPlatform.hpp"
#include "ovpworld.org/memory/ram/1.0/tlm2.0/tlmMemory.hpp"
#include "riscv.ovpworld.org/processor/riscv/1.0/tlm2.0/processor.igen.hpp"

#include "SimpleBus.h"
#include "Sobel.h"

////////////////////////////////////////////////////////////////////////////////
//                                simple Class                                //
////////////////////////////////////////////////////////////////////////////////

class simple {

  public:
    simple (sc_module_name name);

    icmTLMPlatform        Platform;
    SimpleBus 			  <2,3> bus1;
    ram                   ram1;
    ram                   ram2;
    riscv                 cpu1;
    Sobel 				  sobel;
}; // simple

////////////////////////////////////////////////////////////////////////////////
//                              simple Constructor                            //
////////////////////////////////////////////////////////////////////////////////

simple::simple (sc_module_name name)
    : Platform ("icm", ICM_VERBOSE | ICM_STOP_ON_CTRLC| ICM_ENABLE_IMPERAS_INTERCEPTS | ICM_DISASSEMBLE_LABELS)
    , bus1("bus1")
    , ram1 ("ram1", "sp1", 0x80000000)
    , ram2 ("ram2", "sp1", 0x10000)
    , cpu1 (
        "cpu1", //module name
        0,      //processor id
        ICM_ATTR_DEFAULT,  //ICM attributes
        NULL,   //
        "/users/others/jjlab/tools/Imperas.20170511/lib/Linux64/ImperasLib/riscv.ovpworld.org/semihosting/riscv32Newlib/1.0/model.so")
    , sobel ("sobel")
{

    // bus1 masters
    cpu1.INSTRUCTION.socket(bus1.target_socket[0]);
    cpu1.DATA.socket(bus1.target_socket[1]);

    // bus1 slaves
		bus1.bindTargetSocket( ram1.sp1, 0x0, 0x7fffffff, 0x7fffffff );

		bus1.bindTargetSocket( sobel.targ_skt, 0x90000000, 0x9000000f, 0xf );
		
		bus1.bindTargetSocket( ram2.sp1, 0xffff0000, 0xffffffff, 0x0000ffff );
		
}

////////////////////////////////////////////////////////////////////////////////
//                                   M A I N                                  //
////////////////////////////////////////////////////////////////////////////////

int sc_main (int argc, char *argv[] )
{
    const char *exe = "defaultFile";

    if (argc == 2) {
        exe = argv[1];
    } else {
        cout << "Usage: platform.exe <application>" << endl;
        cout << "       Please specify application" << endl;
        return 0;
    }

    cout << "Constructing." << endl;
    simple top("top");
		Uns64 instr_per_sec = 1000000000; //1G
		top.Platform.quantum( sc_time(1000000, SC_NS) ); // 1ms quantum period
		top.bus1.set_clock_period( sc_core::sc_time(1,SC_SEC)/instr_per_sec ); //1ns 

    cout << "Using SystemC TLM2.0 Memory with DMI" << endl;

    top.cpu1.loadLocalMemory(exe, (icmLoaderAttrs)(ICM_LOAD_VERBOSE | ICM_SET_START) );

    cout << "Starting sc_main." << endl;
    sc_core::sc_start();                  // start the simulation
    cout << "Finished sc_main." << endl;
    return 0;                             // return okay status
}

