#include "global.h"
#include "compress.h"
#include "jdct.h"

#include "ts_debug.h"

/* static Variable */
savable_state state_buf;
savable_state *state;

c_derived_tbl* dctbl0_ptr()
{
	static c_derived_tbl dctbl0;
	return &dctbl0;
}
c_derived_tbl* actbl0_ptr()
{
	static c_derived_tbl actbl0;
	return &actbl0;
}
c_derived_tbl* dctbl1_ptr()
{
	static c_derived_tbl dctbl1;
	return &dctbl1;
}
c_derived_tbl* actbl1_ptr()
{
	static c_derived_tbl actbl1;
	return &actbl1;
}
int * dtbl0() 
{
	static int dtbl0[DCTSIZE2];
	return &dtbl0[0];
}
int * dtbl1() 
{
	static int dtbl1[DCTSIZE2];
	return &dtbl1[0];
}

void emit_byte(int val);
void write_file_header();
void write_frame_header();
void write_scan_header();
void write_file_trailer();

void init_htbl(unsigned char *bits, unsigned char *huffval,
  c_derived_tbl *htbl)
{
  int h, i, j;
  int n, v, c;

  for (i = 0; i < 256; i++) {
    htbl->ehufsi[i] = 0;
    htbl->ehufco[i] = 0;
  }
  h = 0;
  c = 0; /* Huffman code */
  for (i = 1; i <= 16; i++) {
    n = bits[i];
    for (j = 0; j < n; j++) {
      if (h >= 256) {
        fprintf(stderr, "%s: bad Huffman table\n", "cjpeg");
        exit(1);
      }
      v = huffval[h++];
      if (v >= 256) {
        fprintf(stderr, "%s: bad Huffman table\n", "cjpeg");
        exit(1);
      }
      htbl->ehufsi[v] = i;
      htbl->ehufco[v] = c++;
    }
    c <<= 1;
  }
}

void init_compress()
{
  int i;

  state = &state_buf;
  state->put_buffer = 0;
  state->put_bits = 0;
  for (i = 0; i < DCTSIZE2; i++) {
    dtbl0()[i] = qtbl0[i] << 3;
    dtbl1()[i] = qtbl1[i] << 3;
  }
  init_htbl(dc_bits0, dc_huffval0, dctbl0_ptr() );
  init_htbl(ac_bits0, ac_huffval0, actbl0_ptr() );
  init_htbl(dc_bits1, dc_huffval1, dctbl1_ptr() );
  init_htbl(ac_bits1, ac_huffval1, actbl1_ptr() );
}

/* from "jchuff.c" */
int emit_bits (unsigned short code, char size)
/* Emit some bits; return TRUE if successful, FALSE if must suspend */
{
	D(
	if ( size > 8 )
		printf ("%4x, %d\n", code, size );
	 )
  /* This routine is heavily used, so it's worth coding tightly. */
  register long put_buffer = (long) code;
  register int put_bits = state->put_bits;

  /* if size is 0, caller used an invalid Huffman table entry */
  if (size == 0) {
    fprintf(stderr, "%s: Huffman missing code\n", "cjpeg");
    exit(1);
  }

  put_buffer &= (((long) 1)<<size) - 1; /* mask off any extra bits in code */
  
  put_bits += size;		   /* new number of bits in buffer */
  
  put_buffer <<= 24 - put_bits;    /* align incoming bits */

  put_buffer |= state->put_buffer; /* and merge with old buffer contents */
  
  while (put_bits >= 8) {
    int c = (int) ((put_buffer >> 16) & 0xFF);
    emit_byte(c);
    if (c == 0xFF) {		   /* need to stuff a zero byte? */
      emit_byte(0);
    }
    put_buffer <<= 8;
    put_bits -= 8;
  }

  state->put_buffer = put_buffer;  /* update state variables */
  state->put_bits = put_bits;

  return TRUE;
}

/* from "jchuff.c" */
int flush_bits ()
{
  if (! emit_bits(0x7F, 7)) /* fill any partial byte with ones */
    return FALSE;
  state->put_buffer = 0;    /* and reset bit-buffer to empty */
  state->put_bits = 0;
  return TRUE;
}

/* from "jchuff.c" */
int encode_one_block(int *block, c_derived_tbl *dctbl, c_derived_tbl *actbl,
                     int last_dc_val)
{
  register int temp, temp2;
  register int nbits;
  register int k, r, i;
  
  /* Encode the DC coefficient difference per section F.1.2.1 */
  temp = temp2 = block[0] - last_dc_val;

  if (temp < 0) {
    temp = -temp;		/* temp is abs value of input */
    /* For a negative input, want temp2 = bitwise complement of abs(input) */
    /* This code assumes we are on a two's complement machine */
    temp2--;
  }
  
  /* Find the number of bits needed for the magnitude of the coefficient */
  nbits = 0;
  while (temp) {
    nbits++;
    temp >>= 1;
  }
  /* Check for out-of-range coefficient values.
   * Since we're encoding a difference, the range limit is twice as much.
   */
  if (nbits > MAX_COEF_BITS+1) {
    fprintf(stderr, "%s: bad DCT coefficient\n", "cjpeg");
    exit(1);
  }
  
  /* Emit the Huffman-coded symbol for the number of bits */
  if (! emit_bits(dctbl->ehufco[nbits], dctbl->ehufsi[nbits]))
    return FALSE;

  /* Emit that number of bits of the value, if positive, */
  /* or the complement of its magnitude, if negative. */
  if (nbits)			/* emit_bits rejects calls with size 0 */
    if (! emit_bits((unsigned int) temp2, nbits))
      return FALSE;

  /* Encode the AC coefficients per section F.1.2.2 */
  
  r = 0;			/* r = run length of zeros */
  
  for (k = 1; k < DCTSIZE2; k++) {
    if ((temp = block[jpeg_natural_order[k]]) == 0) {
      r++;
    } else {
      /* if run length > 15, must emit special run-length-16 codes (0xF0) */
      while (r > 15) {
	if (! emit_bits(actbl->ehufco[0xF0], actbl->ehufsi[0xF0]))
	  return FALSE;
	r -= 16;
      }

      temp2 = temp;
      if (temp < 0) {
	temp = -temp;		/* temp is abs value of input */
	/* This code assumes we are on a two's complement machine */
	temp2--;
      }
      
      /* Find the number of bits needed for the magnitude of the coefficient */
      nbits = 1;		/* there must be at least one 1 bit */
      while ((temp >>= 1))
	nbits++;
      /* Check for out-of-range coefficient values */
      if (nbits > MAX_COEF_BITS) {
        fprintf(stderr, "%s: bad DCT coefficient\n", "cjpeg");
      	exit(1);
      }

      /* Emit Huffman symbol for run length / number of bits */
      i = (r << 4) + nbits;
      if (! emit_bits(actbl->ehufco[i], actbl->ehufsi[i]))
	return FALSE;

      /* Emit that number of bits of the value, if positive, */
      /* or the complement of its magnitude, if negative. */
      if (! emit_bits((unsigned int) temp2, nbits))
	return FALSE;
      
      r = 0;
    }
  }

  /* If the last coef(s) were zero, emit an end-of-block code */
  if (r > 0)
    if (! emit_bits(actbl->ehufco[0], actbl->ehufsi[0]))
      return FALSE;

  return TRUE;
}

void dct_one_block(int *workspace)
//void dct::jpeg_fdct_islow()
{
  INT32 tmp0, tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7;
  INT32 tmp10, tmp11, tmp12, tmp13;
  INT32 z1, z2, z3, z4, z5;
  int i;
  SHIFT_TEMPS
  
  /* Pass 1: process rows. */
  /* Note results are scaled up by sqrt(8) compared to a true DCT; */
  /* furthermore, we scale the results by 2**PASS1_BITS. */

  for ( i = DCTSIZE-1; i >= 0; i--) { /* synopsys line_label pass1_loop */
    tmp0 = workspace[i*DCTSIZE + 0] + workspace[i*DCTSIZE + 7];
    tmp7 = workspace[i*DCTSIZE + 0] - workspace[i*DCTSIZE + 7];
    tmp1 = workspace[i*DCTSIZE + 1] + workspace[i*DCTSIZE + 6];
    tmp6 = workspace[i*DCTSIZE + 1] - workspace[i*DCTSIZE + 6];
    tmp2 = workspace[i*DCTSIZE + 2] + workspace[i*DCTSIZE + 5];
    tmp5 = workspace[i*DCTSIZE + 2] - workspace[i*DCTSIZE + 5];
    tmp3 = workspace[i*DCTSIZE + 3] + workspace[i*DCTSIZE + 4];
    tmp4 = workspace[i*DCTSIZE + 3] - workspace[i*DCTSIZE + 4];
    
    /* Even part per LL&M figure 1 --- note that published figure is faulty;
     * rotator "sqrt(2)*c1" should be "sqrt(2)*c6".
     */
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    
    workspace[i*DCTSIZE + 0] = (DCTELEM) ((tmp10 + tmp11) << PASS1_BITS);
    workspace[i*DCTSIZE + 4] = (DCTELEM) ((tmp10 - tmp11) << PASS1_BITS);

    z1 = MULTIPLY(tmp12 + tmp13, FIX_0_541196100);
    workspace[i*DCTSIZE + 2] = (DCTELEM) DESCALE(z1 + MULTIPLY(tmp13, FIX_0_765366865),
						 CONST_BITS-PASS1_BITS);
    workspace[i*DCTSIZE + 6] = (DCTELEM) DESCALE(z1 + MULTIPLY(tmp12, - FIX_1_847759065),
						 CONST_BITS-PASS1_BITS);
    
    /* Odd part per figure 8 --- note paper omits factor of sqrt(2).
     * cK represents cos(K*pi/16).
     * i0..i3 in the paper are tmp4..tmp7 here.
     */
    
    z1 = tmp4 + tmp7;
    z2 = tmp5 + tmp6;
    z3 = tmp4 + tmp6;
    z4 = tmp5 + tmp7;
    z5 = MULTIPLY(z3 + z4, FIX_1_175875602); /* sqrt(2) * c3 */
    
    tmp4 = MULTIPLY(tmp4, FIX_0_298631336); /* sqrt(2) * (-c1+c3+c5-c7) */
    tmp5 = MULTIPLY(tmp5, FIX_2_053119869); /* sqrt(2) * ( c1+c3-c5+c7) */
    tmp6 = MULTIPLY(tmp6, FIX_3_072711026); /* sqrt(2) * ( c1+c3+c5-c7) */
    tmp7 = MULTIPLY(tmp7, FIX_1_501321110); /* sqrt(2) * ( c1+c3-c5-c7) */
    z1 = MULTIPLY(z1, - FIX_0_899976223); /* sqrt(2) * (c7-c3) */
    z2 = MULTIPLY(z2, - FIX_2_562915447); /* sqrt(2) * (-c1-c3) */
    z3 = MULTIPLY(z3, - FIX_1_961570560); /* sqrt(2) * (-c3-c5) */
    z4 = MULTIPLY(z4, - FIX_0_390180644); /* sqrt(2) * (c5-c3) */
    
    z3 += z5;
    z4 += z5;
    
    workspace[i*DCTSIZE + 7] = (DCTELEM) DESCALE(tmp4 + z1 + z3, CONST_BITS-PASS1_BITS);
    workspace[i*DCTSIZE + 5] = (DCTELEM) DESCALE(tmp5 + z2 + z4, CONST_BITS-PASS1_BITS);
    workspace[i*DCTSIZE + 3] = (DCTELEM) DESCALE(tmp6 + z2 + z3, CONST_BITS-PASS1_BITS);
    workspace[i*DCTSIZE + 1] = (DCTELEM) DESCALE(tmp7 + z1 + z4, CONST_BITS-PASS1_BITS);
  }

  /* Pass 2: process columns.
   * We remove the PASS1_BITS scaling, but leave the results scaled up
   * by an overall factor of 8.
   */

  for ( i = DCTSIZE-1; i >= 0; i--) { /* synopsys line_label pass2_loop */
    tmp0 = workspace[i + DCTSIZE*0] + workspace[i + DCTSIZE*7];
    tmp7 = workspace[i + DCTSIZE*0] - workspace[i + DCTSIZE*7];
    tmp1 = workspace[i + DCTSIZE*1] + workspace[i + DCTSIZE*6];
    tmp6 = workspace[i + DCTSIZE*1] - workspace[i + DCTSIZE*6];
    tmp2 = workspace[i + DCTSIZE*2] + workspace[i + DCTSIZE*5];
    tmp5 = workspace[i + DCTSIZE*2] - workspace[i + DCTSIZE*5];
    tmp3 = workspace[i + DCTSIZE*3] + workspace[i + DCTSIZE*4];
    tmp4 = workspace[i + DCTSIZE*3] - workspace[i + DCTSIZE*4];
    
    /* Even part per LL&M figure 1 --- note that published figure is faulty;
     * rotator "sqrt(2)*c1" should be "sqrt(2)*c6".
     */
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    
    workspace[i + DCTSIZE*0] = (DCTELEM) DESCALE(tmp10 + tmp11, PASS1_BITS);
    workspace[i + DCTSIZE*4] = (DCTELEM) DESCALE(tmp10 - tmp11, PASS1_BITS);
    
    z1 = MULTIPLY(tmp12 + tmp13, FIX_0_541196100);
    workspace[i + DCTSIZE*2] = (DCTELEM) DESCALE(z1 + MULTIPLY(tmp13, FIX_0_765366865),
					   CONST_BITS+PASS1_BITS);
    workspace[i + DCTSIZE*6] = (DCTELEM) DESCALE(z1 + MULTIPLY(tmp12, - FIX_1_847759065),
					   CONST_BITS+PASS1_BITS);
    
    /* Odd part per figure 8 --- note paper omits factor of sqrt(2).
     * cK represents cos(K*pi/16).
     * i0..i3 in the paper are tmp4..tmp7 here.
     */
    
    z1 = tmp4 + tmp7;
    z2 = tmp5 + tmp6;
    z3 = tmp4 + tmp6;
    z4 = tmp5 + tmp7;
    z5 = MULTIPLY(z3 + z4, FIX_1_175875602); /* sqrt(2) * c3 */
    
    tmp4 = MULTIPLY(tmp4, FIX_0_298631336); /* sqrt(2) * (-c1+c3+c5-c7) */
    tmp5 = MULTIPLY(tmp5, FIX_2_053119869); /* sqrt(2) * ( c1+c3-c5+c7) */
    tmp6 = MULTIPLY(tmp6, FIX_3_072711026); /* sqrt(2) * ( c1+c3+c5-c7) */
    tmp7 = MULTIPLY(tmp7, FIX_1_501321110); /* sqrt(2) * ( c1+c3-c5-c7) */
    z1 = MULTIPLY(z1, - FIX_0_899976223); /* sqrt(2) * (c7-c3) */
    z2 = MULTIPLY(z2, - FIX_2_562915447); /* sqrt(2) * (-c1-c3) */
    z3 = MULTIPLY(z3, - FIX_1_961570560); /* sqrt(2) * (-c3-c5) */
    z4 = MULTIPLY(z4, - FIX_0_390180644); /* sqrt(2) * (c5-c3) */
    
    z3 += z5;
    z4 += z5;
    
    workspace[i + DCTSIZE*7] = (DCTELEM) DESCALE(tmp4 + z1 + z3,
						 CONST_BITS+PASS1_BITS);
    workspace[i + DCTSIZE*5] = (DCTELEM) DESCALE(tmp5 + z2 + z4,
						 CONST_BITS+PASS1_BITS);
    workspace[i + DCTSIZE*3] = (DCTELEM) DESCALE(tmp6 + z2 + z3,
						 CONST_BITS+PASS1_BITS);
    workspace[i + DCTSIZE*1] = (DCTELEM) DESCALE(tmp7 + z1 + z4,
						 CONST_BITS+PASS1_BITS);
  }

    
}

void quantize_one_block(int *workspace, int *divisors)
{
  register int temp, qval;
  register int i;

  /* from forward_DCT() in "jcdctmgr.c" */
  for (i = 0; i < DCTSIZE2; i++) {
    qval = divisors[i];
    temp = workspace[i];

    if (temp < 0) {
      temp = -temp;
      temp += qval>>1;	/* for rounding */
      DIVIDE_BY(temp, qval);
      temp = -temp;
    } else {
      temp += qval>>1;	/* for rounding */
      DIVIDE_BY(temp, qval);
    }
    workspace[i] = (short) temp;
  }
}

void sample_one_block( int*  workspace, unsigned char *p, int size, int step )
{
  int ii, jj;
	int index;
	index = 0;
	for (ii = 0; ii < size; ii += step) {
		for (jj = 0; jj < size; jj += step) {
			workspace[index] = *(p+(ii*exp_w+jj)) - CENTERJSAMPLE;
			index++;
		}
	}
}

void print_workspace( int * workspace )
{
	int i, j;
	//printf ("DCTSIZE2=%d, DCTSIZE=%d\n", DCTSIZE2, DCTSIZE );
	for ( i = 0; i< DCTSIZE; i++ ) {
		for ( j = 0; j< DCTSIZE; j++ ) {
			printf( "%5d ", workspace[i*DCTSIZE+j] );
		}
		printf ("\n" );
	}
}
/*
*/
