#define MAIN
#include <stdio.h>
#include <string.h>
#include "global.h"
#include "compress.h"

void read_ppm( FILE *ppm_fp ); 				//ppm file loading, defined in read_ppm.c
void init_compress();
void init_rgb_ycc_convert();

void rgb_ycc_convert();	//rgb_ycc_convert.c
void rgb_ycc_convert_macro_block( int, int ); //rgb_ycc_convert.c
void sample_one_block( int*, unsigned char*, int, int ); //defined in compress.c
void dct_one_block(int *workspace);
void quantize_one_block(int *workspace, int *divisors);
int encode_one_block(int *block, c_derived_tbl *dctbl, c_derived_tbl *actbl, int previous_dc_val);

void write_file_header();
void write_frame_header();
void write_scan_header();
void write_file_trailer();
int flush_bits ();

//access external data struction in compress.c
c_derived_tbl * dctbl0_ptr();
c_derived_tbl * actbl0_ptr();
c_derived_tbl * dctbl1_ptr();
c_derived_tbl * actbl1_ptr();
int * dtbl0();
int * dtbl1();

//void cjpegmin(int argc, char **argv)
int main(int argc, char **argv)
{
	FILE* arg_fp;
	int phony_argc = 1;
	char* phony_argv[16];
	char str_temp[128];
	phony_argv[0] = malloc( strlen( argv[0] ) );
	strcpy( phony_argv[0], argv[0] );
	if ( (arg_fp = fopen( "arg", "r" )) == NULL ) {
		fprintf(stderr, "%s: cannot open the argument file \"arg\"\n", program );
		exit(1);
	}
	while ( fscanf( arg_fp, "%s", str_temp ) != EOF ) {
		phony_argv[phony_argc] = malloc( strlen( str_temp ) );
		strcpy( phony_argv[phony_argc], str_temp );
		//printf( "phony_argv[%d] = %s\n", phony_argc, phony_argv[phony_argc] );
		phony_argc++;
	}
#ifdef CHECK_FLOW
	printf ("begin\n");
#endif
  if (phony_argc != 3) {
    fprintf(stderr, "usage: <exe> <ppm_file> <jpeg_file>\n");
    exit(1);
  }
  program = phony_argv[0];
  if ((ppm_fp = fopen(phony_argv[1], "r")) == NULL) {
    fprintf(stderr, "%s: cannot open \"%s\"\n", program, phony_argv[1]);
    exit(1);
  }
  if ((jpg_fp = fopen(phony_argv[2], "w")) == NULL) {
    fprintf(stderr, "%s: cannot open \"%s\"\n", program, phony_argv[2]);
    exit(1);
  }

  /* Read PPM File */
  read_ppm( ppm_fp );

#ifdef CHECK_FLOW
	printf ("initialization\n");
#endif
  /* Initialize data structure for DCT, Quantize, Huffman Encode */
  int i, j, k;
  int previous_dc_val_y;
  int previous_dc_val_cb;
  int previous_dc_val_cr;
  init_rgb_ycc_convert();
  init_compress();			//prepare DCT table, Huffman table

  previous_dc_val_y = 0;		//inital previous_dc_val = 0
  previous_dc_val_cb = 0;
  previous_dc_val_cr = 0;
	
	//tshsu functional parallelism init begin
	unsigned char * p_list[6] = { &comp_y[0], &comp_y[0], &comp_y[0], &comp_y[0], &comp_cb[0], &comp_cr[0]};
	int p_offset_list[6] = { 0, DCTSIZE, exp_w*DCTSIZE, exp_w*DCTSIZE+DCTSIZE, 0, 0 };
	int size_list[6] = { DCTSIZE, DCTSIZE, DCTSIZE, DCTSIZE, MBSIZE, MBSIZE };
	int step_list[6] = { 1, 1, 1, 1, 2, 2 };
	int * qtbl_list[6] = 
		{ dtbl0(), dtbl0(), dtbl0(), dtbl0(), dtbl1(), dtbl1() };
	c_derived_tbl * dctbl_list[6] = 
		{ dctbl0_ptr(), dctbl0_ptr(), dctbl0_ptr(), dctbl0_ptr(), dctbl1_ptr(), dctbl1_ptr() };
	c_derived_tbl * actbl_list[6] = 
		{ actbl0_ptr(), actbl0_ptr(), actbl0_ptr(), actbl0_ptr(), actbl1_ptr(), actbl1_ptr() };
	int * previous_dc_val_ptr_list[6] = { &previous_dc_val_y, &previous_dc_val_y, &previous_dc_val_y, 
		&previous_dc_val_y, &previous_dc_val_cb, &previous_dc_val_cr };
	int workspace[DCTSIZE2];
	//tshsu functional init end

	//write headers
#ifdef CHECK_FLOW
	printf ("write header\n");
#endif
  write_file_header();
  write_frame_header();
  write_scan_header();

#ifdef CHECK_FLOW
	printf ("start\n");
#endif
	for (i = 0; i < exp_h; i += MBSIZE) {			//MBSIZE is 16, DCTSIZE is 8, DCTSIZE2 is 64
    for (j = 0; j < exp_w; j += MBSIZE) {

			/* Color Convert */
			/* RGB to YCbCr conversion
			 * input comp_r/g/b, output comp_y/cb/cr
			 */
			rgb_ycc_convert_macro_block( i, j );
			
			for (k = 0; k<6; k++ ) { 

				//stage 1: sampling
				sample_one_block
					( workspace, p_list[k]+(i*exp_w+j)+p_offset_list[k], size_list[k], step_list[k] );

				/*
				printf( "1. sampleing: %d, %d, %d:", i, j, k );
				print_workspace( workspace );
				*/

				//stage 2: dct
				dct_one_block(workspace);
				
				/*
				printf( "2. dct: %d, %d, %d:", i, j, k );
				print_workspace( workspace );
				*/

				//stage 3: quantize
				quantize_one_block(workspace, qtbl_list[k]);
				
				/*
				printf( "3. quantize: %d, %d, %d:", i, j, k );
				print_workspace( workspace );
				*/

				//stage 4: Huffman + output
				encode_one_block(workspace, dctbl_list[k], actbl_list[k], *previous_dc_val_ptr_list[k]);
				*previous_dc_val_ptr_list[k] = workspace[0];
				
				/*
				printf( "4. Huffman: %d, %d, %d:", i, j, k );
				print_workspace( workspace );
				*/

			}
    }
  }

  flush_bits();

#ifdef CHECK_FLOW
	printf ("write trailer\n");
#endif
  write_file_trailer();

	fclose( ppm_fp );
	fclose( jpg_fp );
#ifdef CHECK_FLOW
	printf( "end\n" );
#endif
  return 0;
}
