#include "global.h"
#include "jccolor.h"

/* static Variable */
static long rgb_ycc_tab[TABLE_SIZE];

//to be modified
void init_rgb_ycc_convert()
{
  int i;

  /* from rgb_ycc_start() in "jccolor.c" */
  for (i = 0; i <= MAXJSAMPLE; i++) {
    rgb_ycc_tab[i+R_Y_OFF] = FIX(0.29900) * i;
    rgb_ycc_tab[i+G_Y_OFF] = FIX(0.58700) * i;
    rgb_ycc_tab[i+B_Y_OFF] = FIX(0.11400) * i     + ONE_HALF;
    rgb_ycc_tab[i+R_CB_OFF] = (-FIX(0.16874)) * i;
    rgb_ycc_tab[i+G_CB_OFF] = (-FIX(0.33126)) * i;
    /* We use a rounding fudge-factor of 0.5-epsilon for Cb and Cr.
     * This ensures that the maximum output will round to MAXJSAMPLE
     * not MAXJSAMPLE+1, and thus that we don't have to range-limit.
     */
    rgb_ycc_tab[i+B_CB_OFF] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
/*  B=>Cb and R=>Cr tables are the same
    rgb_ycc_tab[i+R_CR_OFF] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
*/
    rgb_ycc_tab[i+G_CR_OFF] = (-FIX(0.41869)) * i;
    rgb_ycc_tab[i+B_CR_OFF] = (-FIX(0.08131)) * i;
  }
}

unsigned char rgb_to_y(unsigned char r, unsigned char g, unsigned char b)
{
  /* Y */
  unsigned char y;
  long *ctab = rgb_ycc_tab;
  y = (JSAMPLE)
    ((ctab[r+R_Y_OFF] + ctab[g+G_Y_OFF] + ctab[b+B_Y_OFF])
     >> SCALEBITS);
  return(y);
}

unsigned char rgb_to_cb(unsigned char r, unsigned char g, unsigned char b)
{
  /* Cb */
  unsigned char cb;
  long *ctab = rgb_ycc_tab;
  cb = (JSAMPLE)
    ((ctab[r+R_CB_OFF] + ctab[g+G_CB_OFF] + ctab[b+B_CB_OFF])
     >> SCALEBITS);
  return(cb);
}

unsigned char rgb_to_cr(unsigned char r, unsigned char g, unsigned char b)
{
  /* Cr */
  unsigned char cr;
  long *ctab = rgb_ycc_tab;
  cr = (JSAMPLE)
    ((ctab[r+R_CR_OFF] + ctab[g+G_CR_OFF] + ctab[b+B_CR_OFF])
     >> SCALEBITS);
  return(cr);
}

void rgb_ycc_convert()
{
  int i, j;
  unsigned char r, g, b;
  unsigned char *r_ptr, *g_ptr, *b_ptr;
  unsigned char *y_ptr, *cb_ptr, *cr_ptr;

  init_rgb_ycc_convert();
  r_ptr = &comp_r[0];
  g_ptr = &comp_g[0];
  b_ptr = &comp_b[0];
  y_ptr = &comp_y[0];
  cb_ptr = &comp_cb[0];
  cr_ptr = &comp_cr[0];
  for (i = 0; i < org_h; i++){
    for (j = 0; j < exp_w; j++){
      r = (j < org_w) ? *r_ptr++ : *(r_ptr-1); /* expand right edge */
      g = (j < org_w) ? *g_ptr++ : *(g_ptr-1);
      b = (j < org_w) ? *b_ptr++ : *(b_ptr-1);

      *y_ptr++ = rgb_to_y(r, g, b);
      *cb_ptr++ = rgb_to_cb(r, g, b);
      *cr_ptr++ = rgb_to_cr(r, g, b);
    }
  }
  /* expand lower edge */
  for (i = org_h; i < exp_h; i++) {
    for (j = 0; j < exp_w; j++) {
      *y_ptr++ = *(y_ptr - exp_w);
      *cb_ptr++ = *(cb_ptr - exp_w);
      *cr_ptr++ = *(cr_ptr - exp_w);
    }
  }
}

void rgb_ycc_convert_pixel( 
		unsigned char *y_ptr,
		unsigned char *cb_ptr,
		unsigned char *cr_ptr,
		unsigned char r,
		unsigned char g,
		unsigned char b
		)
{
		*y_ptr = rgb_to_y(r, g, b);
		*cb_ptr = rgb_to_cb(r, g, b);
		*cr_ptr = rgb_to_cr(r, g, b);
}

void rgb_ycc_convert_macro_block( int y, int x )
{
  int i, j;
  unsigned char r, g, b;
  unsigned char *r_ptr, *g_ptr, *b_ptr;
  unsigned char *y_ptr, *cb_ptr, *cr_ptr;

  //init_rgb_ycc_convert();
  r_ptr = &comp_r[y*org_w+x];
  g_ptr = &comp_g[y*org_w+x];
  b_ptr = &comp_b[y*org_w+x];
  y_ptr = &comp_y[y*exp_w+x];
  cb_ptr = &comp_cb[y*exp_w+x];
  cr_ptr = &comp_cr[y*exp_w+x];
  for (i = 0; i < MBSIZE; i++){
    for (j = 0; j < MBSIZE; j++){
      r = *(r_ptr+(i*org_w+j)); 
      g = *(g_ptr+(i*org_w+j));
      b = *(b_ptr+(i*org_w+j));
			//
			if ( x+j >= org_w || y+i >= org_h ) {
				if ( x+j >= org_w && y+i >= org_h ) {
					r = comp_r[ org_h * org_w - 1 ];
					g = comp_g[ org_h * org_w - 1 ];
					b = comp_b[ org_h * org_w - 1 ];
				}
				else if ( x+j >= org_w ) {
					r = comp_r[ (y+i) * org_w + org_w - 1 ];
					g = comp_g[ (y+i) * org_w + org_w - 1 ];
					b = comp_b[ (y+i) * org_w + org_w - 1 ];
				}
				else if ( y+i >= org_h ) {
					r = comp_r[ org_h * org_w - org_w + x+j ];
					g = comp_g[ org_h * org_w - org_w + x+j ];
					b = comp_b[ org_h * org_w - org_w + x+j ];
				}
			}
			/* mudify the interface */
			rgb_ycc_convert_pixel ( (y_ptr+(i*exp_w+j)), (cb_ptr+(i*exp_w+j)), (cr_ptr+(i*exp_w+j)), r, g, b );
    }
  }
}
