package sobel
import chisel3._
import chisel3.util._
import config.{Parameters, Field}
import uncore.tilelink._
import uncore.tilelink2._
import junctions._
import diplomacy._
import regmapper.{HasRegMap, RegField}
import rocketchip._
import _root_.util.UIntIsOneOf

case class SobelParams(address: BigInt, beatBytes: Int)

/*
class SobelBase(w: Int) extends Module {
  val io = IO(new Bundle {

    val a  = Input(UInt(w.W))
    val b  = Input(UInt(w.W))
    val e  = Input(Bool())
    val v  = Output(Bool())
    val z  = Output(UInt(w.W))

  })

  // Sobel
  val x = Reg(UInt(w.W))
  val y = Reg(UInt(w.W))

  when   (x > y) {
    x := x - y
  }
    .elsewhen (x <= y) { y := y - x }

  when (io.e) { x := io.a; y := io.b }
  io.z := x
  io.v := y === 0.U

}
*/

/*
class Mux2 extends Module {
  val io = IO(new Bundle {
    val sel = Input(UInt(1.W))
    val in0 = Input(UInt(32.W))
    val in1 = Input(UInt(32.W))
    val out = Output(UInt(32.W))
  })
  when(io.sel === 0.U) { io.out := io.in0 }
  when(io.sel === 1.U) { io.out := io.in1 }

}
*/


class SobelBase(w: Int) extends Module {
  val io = IO(new Bundle {
    val e = Input(Bool())
    val m0 = Input(UInt(w.W))
    val m1 = Input(UInt(w.W))
    val m2 = Input(UInt(w.W))
    val m3 = Input(UInt(w.W))
    // val m4 = Input(UInt(w.W))
    val m5 = Input(UInt(w.W))
    val m6 = Input(UInt(w.W))
    val m7 = Input(UInt(w.W))
    val m8 = Input(UInt(w.W))
    val v = Output(Bool())
    val out_X = Output(UInt(w.W))
    val out_Y = Output(UInt(w.W))
//    val square = Output(UInt(w.W))
  })

  val cnt = Reg(UInt(3.W))
  when (io.e) { cnt := 0.U }
  .elsewhen (cnt === 3.U) { cnt := 3.U }
  .otherwise { cnt := cnt + 1.U }

  
  val r0 = Reg(UInt(w.W))
  val r1 = Reg(UInt(w.W))
  val r2 = Reg(UInt(w.W))
  val r3 = Reg(UInt(w.W))
  val r5 = Reg(UInt(w.W))
  val r6 = Reg(UInt(w.W))
  val r7 = Reg(UInt(w.W))
  val r8 = Reg(UInt(w.W))

  when (io.e) {
    r0 := io.m0
    r1 := io.m1
    r2 := io.m2
    r3 := io.m3
    r5 := io.m5
    r6 := io.m6
    r7 := io.m7
    r8 := io.m8
  }
/*
  val X_dir = Module(new Mux2())
  // val row_0 = io.m0 + io.m1 + io.m1 + io.m2
  // val row_2 = io.m6 + io.m7 + io.m7 + io.m8
  val row_0 = r0 + r1 + r1 + r2
  val row_2 = r6 + r7 + r7 + r8
  X_dir.io.sel := Mux(row_0 > row_2, 1.U, 0.U)
  X_dir.io.in0 := row_2 - row_0
  X_dir.io.in1 := row_0 - row_2

  val Y_dir = Module(new Mux2())
  val col_0 = r0 + r3 + r3 + r6
  val col_2 = r2 + r5 + r5 + r8
  Y_dir.io.sel := Mux(col_0 > col_2, 1.U, 0.U)
  Y_dir.io.in0 := col_2 - col_0
  Y_dir.io.in1 := col_0 - col_2
*/

  val row_0 = r0 + r1 + r1 + r2
  val row_2 = r6 + r7 + r7 + r8
  val col_0 = r0 + r3 + r3 + r6
  val col_2 = r2 + r5 + r5 + r8
  
  val r_x = Reg(UInt(w.W))
  val r_y = Reg(UInt(w.W))

  when (cnt === 2.U) { 
	r_x := Mux(row_0 > row_2, row_0 - row_2, row_2 - row_0)
	r_y := Mux(col_0 > col_2, col_0 - col_2, col_2 - col_0)
  }

  io.out_X := r_x
  io.out_Y := r_y
  io.v := cnt === 3.U
//   io.square := X_dir.io.out * X_dir.io.out + Y_dir.io.out * Y_dir.io.out
}

/*
trait SobelTLBundle extends Bundle {
  val v = Output(Bool())
  val z = Output(UInt(32.W))
}
*/


trait SobelTLBundle extends Bundle {
  val v = Output(Bool())
  val out_X = Output(UInt(32.W))
  val out_Y = Output(UInt(32.W))
//   val square = Output(UInt(32.W))
}

trait SobelTLModule extends Module with HasRegMap {
  val io: SobelTLBundle
  implicit val p: Parameters
  def params: SobelParams

  val w = params.beatBytes * 8
  require(w <= 32)

  // Sobel
  /*
  val e = RegInit(false.B)
  val a = Reg(UInt(w.W))
  val b = Reg(UInt(w.W))
  */

  val e = RegInit(false.B)
  val m0 = Reg(UInt(w.W))
  val m1 = Reg(UInt(w.W))
  val m2 = Reg(UInt(w.W))
  val m3 = Reg(UInt(w.W))
  // val m4 = Reg(UInt(w.W))
  val m5 = Reg(UInt(w.W))
  val m6 = Reg(UInt(w.W))
  val m7 = Reg(UInt(w.W))
  val m8 = Reg(UInt(w.W))

  val base = Module(new SobelBase(w))

  io.v := base.io.v
  io.out_X := base.io.out_X
  io.out_Y := base.io.out_Y
//   io.square := base.io.square
  base.io.e := e
  base.io.m0 := m0
  base.io.m1 := m1
  base.io.m2 := m2
  base.io.m3 := m3
  base.io.m5 := m5
  base.io.m6 := m6
  base.io.m7 := m7
  base.io.m8 := m8

  regmap(
    0x00 -> Seq(
      RegField(w, m0)),
    0x04 -> Seq(
      RegField(w, m1)),
    0x08 -> Seq(
      RegField(w, m2)),
    0x0c -> Seq(
      RegField(w, m3)),
    0x10 -> Seq(
      RegField(w, m5)),
    0x14 -> Seq(
      RegField(w, m6)),
    0x18 -> Seq(
      RegField(w, m7)),
    0x1c -> Seq(
      RegField(w, m8)),
    0x20 -> Seq(
      RegField(w, io.out_X)),
    0x24 -> Seq(
      RegField(w, io.out_Y)),
    0x28 -> Seq(
      RegField(1, io.v)),
    0x2c -> Seq(
      RegField(1, e)))
//     0x28 -> Seq(
//       RegField(w, io.square)))
}

class SobelTL(c: SobelParams)(implicit p: Parameters)
  extends TLRegisterRouter(
    c.address, "sobel", Seq("ucbbar,sobel"),
    beatBytes = c.beatBytes)(
      new TLRegBundle(c, _) with SobelTLBundle)(
      new TLRegModule(c, _, _) with SobelTLModule)

trait PeripherySobel extends LazyModule with HasPeripheryParameters {
  implicit val p: Parameters
  val peripheryBus: TLXbar

  private val address = 0x2000

  val sobel = LazyModule(new SobelTL(
    SobelParams(address, peripheryBusConfig.beatBytes))(p))

  sobel.node := TLFragmenter(
    peripheryBusConfig.beatBytes, cacheBlockBytes)(peripheryBus.node)
}

trait PeripherySobelBundle {
  val v = Output(Bool())
  val out_X = Output(UInt(32.W))
  val out_Y = Output(UInt(32.W))
//   val square = Output(UInt(32.W))
}

trait PeripherySobelModule extends HasPeripheryParameters {
  implicit val p: Parameters
  val io: PeripherySobelBundle
  val outer: PeripherySobel
  io.v := outer.sobel.module.io.v
  io.out_X := outer.sobel.module.io.out_X
  io.out_Y := outer.sobel.module.io.out_Y
//   io.square := outer.sobel.module.io.square
}
