#include <stdio.h>
#include <stdlib.h>
#include "hardware-thread.h"

/*
 *  Use Gregory-Leibniz Series to calculate Pi:
 *
 *    Pi/4 = 1 - (1/3) + (1/5) - (1/7) + ....
 *
 */

double pi_over_4 = 0;
Mutex mutex;

int pi_check(const int num_terms, const double pi) {
  double golden_pi_over4 = 0;
  double factor = 1;
  int i;
  for (i = 0; i < num_terms; i++, factor = -factor) {
    golden_pi_over4 += factor / (2 * i + 1);
  }
  double golden_pi = golden_pi_over4 * 4.0;
  if ( (golden_pi-pi) < 0.000001)  {
    printf("Pass\n");
  } else {
    printf("Expected Pi: %f, got %f\n", golden_pi, pi);
    printf("Fail\n");
  }
}

void thread_entry(int cid, int nc) {
  unsigned total_terms = 300;

  unsigned num_terms_per_thread = (total_terms + nc - 1) / nc;
  unsigned begin_idx = num_terms_per_thread * cid;
  unsigned end_idx;
  if (cid == nc - 1) {
    end_idx = total_terms;
  } else {
    end_idx = begin_idx + num_terms_per_thread;
  }

  double factor = begin_idx % 2 ? -1.0 : 1.0;

  double local_sum = 0, i;
  for (i = begin_idx; i < end_idx; i++, factor = -factor) {
    local_sum += factor / (2 * i + 1);
  }

  mutex_lock(&mutex);
  pi_over_4 += local_sum;
  mutex_unlock(&mutex);

  barrier(nc);

  if (cid == 0) {
    printf("Pi = %f\n", (pi_over_4) * 4);
    pi_check(total_terms, pi_over_4 * 4);
  }

  exit(0);
}
