// See LICENSE for license details.

package example

import chisel3._
import scala.collection.mutable.ArrayBuffer

class median5 extends Module {
  val io = IO(new Bundle {
    val x0  = Input(UInt(8.W))
    val x1  = Input(UInt(8.W))
    val x2  = Input(UInt(8.W))
    val x3  = Input(UInt(8.W))
    val x4  = Input(UInt(8.W))
    val cnt = Input(UInt(5.W))
    val e   = Input(Bool())
    val z   = Output(UInt(8.W))
  })

  val tbl = Reg(Vec(5,UInt(8.W)))
  when (io.e) {
    tbl(0) := io.x0
    tbl(1) := io.x1
    tbl(2) := io.x2
    tbl(3) := io.x3
    tbl(4) := io.x4
  }

  for (i <- 0 until 5) {
    for (j <- 1 until 5) {
      when (io.cnt === (4*i + j).asUInt(5.W)) {
        when (tbl(j) < tbl(j-1)) {
          tbl(j) := tbl(j-1)
          tbl(j-1) := tbl(j)
        }
      }
    }
  }

  io.z := tbl(2)
}

class median3 extends Module {
  val io = IO(new Bundle {
    val x0  = Input(UInt(8.W))
    val x1  = Input(UInt(8.W))
    val x2  = Input(UInt(8.W))
    val cnt = Input(UInt(5.W))
    val cnt_1 = Input(UInt(5.W))
    val z   = Output(UInt(8.W))
  })

  val tbl = Reg(Vec(3,UInt(8.W)))
  when (io.cnt === 15.U) {
    tbl(0) := io.x0
    tbl(1) := io.x1
    tbl(2) := io.x2
  }
  for (i <- 0 until 3) {
    for (j <- 1 until 3) {
      when (io.cnt_1 === (2*i + j).asUInt(5.W)) {
        when (tbl(j) < tbl(j-1)) {
          tbl(j) := tbl(j-1)
          tbl(j-1) := tbl(j)
        }
      }
    }
  }

  io.z := tbl(1)
}

class hybrid extends Module {
  val io = IO(new Bundle {
    val R0 = Input(UInt(8.W))
    val R1 = Input(UInt(8.W))
    val R2 = Input(UInt(8.W))
    val R3 = Input(UInt(8.W))
    val R4 = Input(UInt(8.W))
    val R5 = Input(UInt(8.W))
    val R6 = Input(UInt(8.W))
    val R7 = Input(UInt(8.W))
    val R8 = Input(UInt(8.W))

    val G0 = Input(UInt(8.W))
    val G1 = Input(UInt(8.W))
    val G2 = Input(UInt(8.W))
    val G3 = Input(UInt(8.W))
    val G4 = Input(UInt(8.W))
    val G5 = Input(UInt(8.W))
    val G6 = Input(UInt(8.W))
    val G7 = Input(UInt(8.W))
    val G8 = Input(UInt(8.W))

    val B0 = Input(UInt(8.W))
    val B1 = Input(UInt(8.W))
    val B2 = Input(UInt(8.W))
    val B3 = Input(UInt(8.W))
    val B4 = Input(UInt(8.W))
    val B5 = Input(UInt(8.W))
    val B6 = Input(UInt(8.W))
    val B7 = Input(UInt(8.W))
    val B8 = Input(UInt(8.W))

    val e  = Input(Bool())
    val ansR = Output(UInt(8.W))
    val ansG = Output(UInt(8.W))
    val ansB = Output(UInt(8.W))
  })

  val cnt = RegInit(1.U(5.W))
  when(cnt < 16.U) {
    cnt := cnt + 1.U
  } .elsewhen (io.e) {
    cnt := 1.U
  }

  val cnt_1 = RegInit(1.U(5.W))
  when(cnt === 15.U) {
    cnt_1 := cnt_1 + 1.U
  } .elsewhen (io.e || cnt_1 === 6.U) {
    cnt_1 := 1.U
  } .elsewhen (cnt_1 > 1.U) {
    cnt_1 := cnt_1 + 1.U
  }

  val R_0 = Module(new median5())
  R_0.io.x0  := io.R1
  R_0.io.x1  := io.R3
  R_0.io.x2  := io.R4
  R_0.io.x3  := io.R5
  R_0.io.x4  := io.R7
  R_0.io.cnt := cnt
  R_0.io.e   := io.e

  val R_1 = Module(new median5())
  R_1.io.x0  := io.R0
  R_1.io.x1  := io.R2
  R_1.io.x2  := io.R4
  R_1.io.x3  := io.R6
  R_1.io.x4  := io.R8
  R_1.io.cnt := cnt
  R_1.io.e   := io.e

  val R_2 = Module(new median3())
  R_2.io.x0  := R_0.io.z
  R_2.io.x1  := R_1.io.z
  R_2.io.x2  := io.R4
  R_2.io.cnt := cnt
  R_2.io.cnt_1 := cnt_1

  val G_0 = Module(new median5())
  G_0.io.x0  := io.G1
  G_0.io.x1  := io.G3
  G_0.io.x2  := io.G4
  G_0.io.x3  := io.G5
  G_0.io.x4  := io.G7
  G_0.io.cnt := cnt
  G_0.io.e   := io.e

  val G_1 = Module(new median5())
  G_1.io.x0  := io.G0
  G_1.io.x1  := io.G2
  G_1.io.x2  := io.G4
  G_1.io.x3  := io.G6
  G_1.io.x4  := io.G8
  G_1.io.cnt := cnt
  G_1.io.e   := io.e

  val G_2 = Module(new median3())
  G_2.io.x0  := G_0.io.z
  G_2.io.x1  := G_1.io.z
  G_2.io.x2  := io.G4
  G_2.io.cnt := cnt
  G_2.io.cnt_1 := cnt_1

  val B_0 = Module(new median5())
  B_0.io.x0  := io.B1
  B_0.io.x1  := io.B3
  B_0.io.x2  := io.B4
  B_0.io.x3  := io.B5
  B_0.io.x4  := io.B7
  B_0.io.cnt := cnt
  B_0.io.e   := io.e

  val B_1 = Module(new median5())
  B_1.io.x0  := io.B0
  B_1.io.x1  := io.B2
  B_1.io.x2  := io.B4
  B_1.io.x3  := io.B6
  B_1.io.x4  := io.B8
  B_1.io.cnt := cnt
  B_1.io.e   := io.e

  val B_2 = Module(new median3())
  B_2.io.x0  := B_0.io.z
  B_2.io.x1  := B_1.io.z
  B_2.io.x2  := io.B4
  B_2.io.cnt := cnt
  B_2.io.cnt_1 := cnt_1

  io.ansR := R_2.io.z
  io.ansG := G_2.io.z
  io.ansB := B_2.io.z
}
