#ifndef _HYBRID_H
#define _HYBRID_H

#include <iostream>
#include <iomanip>
#include <sstream>

#include <systemc>
#include "tlm_utils/simple_target_socket.h"
#include "tlm_utils/simple_initiator_socket.h"

// #include "hybrid_tlm.h"

// Caution: The verilator related headers must include after SystemC headers.
//          Otherwise it will finds no SYSTEMC_VERSION and assume you
//          are using oldest SystemC API.
#include "Vhybrid.h"

using namespace sc_core;
using namespace tlm;


typedef union word {
    int sint;
    unsigned int uint;
    unsigned char uc[4];
} word;


SC_MODULE(Hybrid)
{
    tlm_utils::simple_target_socket<Hybrid> targ_skt;

    SC_CTOR(Hybrid):
        targ_skt("hybrid_target_socket"), m_clk_period(sc_time(1, SC_NS))
    {
        // reset phase
        m_hybrid.reset = 1;
        // make a rising edge to clock
        step();
        // end of reset phase
        m_hybrid.reset = 0;
        // register the TLM transport function
        targ_skt.register_b_transport(this, &Hybrid::b_transport);
        SC_THREAD(Hybrid_process);
    }

    void step() {
        // make a rising edge to clock
        m_hybrid.clock = 0;
        m_hybrid.eval();
        m_hybrid.clock = 1;
        m_hybrid.eval();
    }

    void Hybrid_process()
    {
        while (true)
        {
            unsigned int cycle = 0;

            word package;

            package.uint = num_fifo.read();
            m_hybrid.io_R0 = package.uc[3];
            m_hybrid.io_R1 = package.uc[2];
            m_hybrid.io_R2 = package.uc[1];

            package.uint = num_fifo.read();
            m_hybrid.io_G0 = package.uc[3];
            m_hybrid.io_G1 = package.uc[2];
            m_hybrid.io_G2 = package.uc[1];

            package.uint = num_fifo.read();
            m_hybrid.io_B0 = package.uc[3];
            m_hybrid.io_B1 = package.uc[2];
            m_hybrid.io_B2 = package.uc[1];

            package.uint = num_fifo.read();
            m_hybrid.io_R3 = package.uc[3];
            m_hybrid.io_R4 = package.uc[2];
            m_hybrid.io_R5 = package.uc[1];

            package.uint = num_fifo.read();
            m_hybrid.io_G3 = package.uc[3];
            m_hybrid.io_G4 = package.uc[2];
            m_hybrid.io_G5 = package.uc[1];

            package.uint = num_fifo.read();
            m_hybrid.io_B3 = package.uc[3];
            m_hybrid.io_B4 = package.uc[2];
            m_hybrid.io_B5 = package.uc[1];

            package.uint = num_fifo.read();
            m_hybrid.io_R6 = package.uc[3];
            m_hybrid.io_R7 = package.uc[2];
            m_hybrid.io_R8 = package.uc[1];

            package.uint = num_fifo.read();
            m_hybrid.io_G6 = package.uc[3];
            m_hybrid.io_G7 = package.uc[2];
            m_hybrid.io_G8 = package.uc[1];

            package.uint = num_fifo.read();
            m_hybrid.io_B6 = package.uc[3];
            m_hybrid.io_B7 = package.uc[2];
            m_hybrid.io_B8 = package.uc[1];

            m_hybrid.io_e = 1;
            step();
            cycle++;
            m_hybrid.io_e = 0;

            for (int x = 0; x < 20; x++) {
                step();
                cycle++;
            }
            wait(cycle * m_clk_period);

            word output_buffer;
            output_buffer.uc[3] = m_hybrid.io_ansR;
            output_buffer.uc[2] = m_hybrid.io_ansG;
            output_buffer.uc[1] = m_hybrid.io_ansB;
            result_fifo.write(output_buffer.uint);
        }
    }

    // Parse the payload and read from the FIFO or write to the FIFO
    void b_transport( tlm_generic_payload & trans, sc_time & delay )
    {
        tlm_command cmd = trans.get_command();
        // unsigned *       ptr = reinterpret_cast<unsigned *>(trans.get_data_ptr());
        // unsigned int     len = trans.get_data_length();
        unsigned char *data_ptr = trans.get_data_ptr();

        word in_package;
        word out_package;
        switch (cmd) {
        case TLM_WRITE_COMMAND:
            // num_fifo.write(*ptr);
            in_package.uc[3] = data_ptr[3];
            in_package.uc[2] = data_ptr[2];
            in_package.uc[1] = data_ptr[1];
            in_package.uc[0] = data_ptr[0];
            num_fifo.write(in_package.uint);
            wait(1 * m_clk_period);
            break;

        case TLM_READ_COMMAND:
            out_package.uint = result_fifo.read();
            data_ptr[3] = out_package.uc[3];
            data_ptr[2] = out_package.uc[2];
            data_ptr[1] = out_package.uc[1];
            data_ptr[0] = out_package.uc[0];
            break;

        default:
            break;
        }

        trans.set_response_status( TLM_OK_RESPONSE );
    }

private:
    Vhybrid m_hybrid;
    sc_fifo<unsigned int> num_fifo;
    sc_fifo<unsigned int> result_fifo;
    const sc_time m_clk_period;
};

#endif
