#define MAIN
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define MASK_N 2
#define MASK_X 3
#define MASK_Y 3
#define WHITE  255
#define BLACK  0
#define HEIGHT 361
#define WIDTH 480
#define BIT_PER_PIXEL 24

unsigned char image_s_arr[HEIGHT * WIDTH * BIT_PER_PIXEL] = {"0"};
unsigned char image_t_arr[HEIGHT * WIDTH * BIT_PER_PIXEL] = {"0"};
FILE *fp_s = NULL;                 // source file handler
FILE *fp_t = NULL;                 // target file handler

unsigned int   width, height;      // image width, image height
unsigned int   rgb_raw_data_offset;// RGB raw data offset
unsigned char  bit_per_pixel;      // bit per pixel
unsigned short byte_per_pixel;     // byte per pixel

// bitmap header
unsigned char header[54] =
{
    0x42,        // identity : B
    0x4d,        // identity : M
    0, 0, 0, 0,  // file size
    0, 0,        // reserved1
    0, 0,        // reserved2
    54, 0, 0, 0, // RGB data offset
    40, 0, 0, 0, // struct BITMAPINFOHEADER size
    0, 0, 0, 0,  // bmp width
    0, 0, 0, 0,  // bmp height
    1, 0,        // planes
    24, 0,       // bit per pixel
    0, 0, 0, 0,  // compression
    0, 0, 0, 0,  // data size
    0, 0, 0, 0,  // h resolution
    0, 0, 0, 0,  // v resolution
    0, 0, 0, 0,  // used colors
    0, 0, 0, 0   // important colors
};

int read_bmp(const char *fname_s)
{
    fp_s = fopen(fname_s, "rb");
    if (fp_s == NULL)
    {
        printf("fopen fp_s error\n");
        return -1;
    }

    // move offset to 10 to find rgb raw data offset
    fseek(fp_s, 10, SEEK_SET);
    fread(&rgb_raw_data_offset, sizeof(unsigned int), 1, fp_s);

    // move offset to 18 to get width & height;
    fseek(fp_s, 18, SEEK_SET);
    fread(&width,  sizeof(unsigned int), 1, fp_s);
    fread(&height, sizeof(unsigned int), 1, fp_s);

    // get bit per pixel
    fseek(fp_s, 28, SEEK_SET);
    fread(&bit_per_pixel, sizeof(unsigned short), 1, fp_s);
    byte_per_pixel = bit_per_pixel / 8;
    printf("bit_per_pixel: %d\n", int(bit_per_pixel));
    printf("byte_per_pixel: %d\n", int(byte_per_pixel));
    // check image parameter bound
    if (height > HEIGHT || width > WIDTH || bit_per_pixel > BIT_PER_PIXEL)
        return -1;

    // move offset to rgb_raw_data_offset to get RGB raw data
    fseek(fp_s, rgb_raw_data_offset, SEEK_SET);

    fread(image_s_arr, sizeof(unsigned char), (size_t)(long)width * height * byte_per_pixel, fp_s);

    return 0;
}

// convert RGB to gray level int
//int color_to_int(int r, int g, int b)
//{
//    return (r + g + b) / 3;
//}

unsigned char Hybrid_Median(unsigned char* arr)
{
    unsigned char m0[5] = {arr[1], arr[3], arr[4], arr[5], arr[7]};
    unsigned char m1[5] = {arr[0], arr[2], arr[4], arr[6], arr[8]};
    for (int i = 0; i < 5; i++)
    {
        for (int j = 1; j < 5; j++)
        {
            if(m0[j] < m0[j-1])
            {
                unsigned char tmp = m0[j];
                m0[j] = m0[j-1];
                m0[j-1] = tmp;
            }
        }
    }

    for (int i = 0; i < 5; i++)
    {
        for (int j = 1; j < 5; j++)
        {
            if(m1[j] < m1[j-1])
            {
                unsigned char tmp = m1[j];
                m1[j] = m1[j-1];
                m1[j-1] = tmp;
            }
        }
    }

    unsigned char m2[3] = {m0[2], m1[2], arr[4]};
    for (int i = 0; i < 3; i++)
    {
        for (int j = 1; j < 3; j++)
        {
            if(m2[j] < m2[j-1])
            {
                unsigned char tmp = m2[j];
                m2[j] = m2[j-1];
                m2[j-1] = tmp;
            }
        }
    }

    return m2[1];
}

int sobel()
{
    unsigned int  x, y, i, v, u;             // for loop counter
    unsigned char R, G, B;         // color of R, G, B
    for (y = 0; y < height; y++ )
    {
        for (x = 0; x < width; x++ )
        {
            int cnt = 0;
            unsigned char valR[9] = {0};
            unsigned char valG[9] = {0};
            unsigned char valB[9] = {0};
            for (v = -1; v != 2; ++v)
            {
                for (u = -1; u != 2; ++u)
                {
                    if (x + u >= 0 && x + u < width && y + v >= 0 && y + v < height)
                    {
                        R = *(image_s_arr + byte_per_pixel * (width * (y + v) + (x + u)) + 2);
                        G = *(image_s_arr + byte_per_pixel * (width * (y + v) + (x + u)) + 1);
                        B = *(image_s_arr + byte_per_pixel * (width * (y + v) + (x + u)) + 0);
//                        val[cnt++] = color_to_int(R, G, B);
                        valR[cnt] = R;
                        valG[cnt] = G;
                        valB[cnt] = B;

                    }
                    cnt = cnt + 1;
                }
            }
            *(image_t_arr + byte_per_pixel * (width * y + x) + 2) = Hybrid_Median(valR);
            *(image_t_arr + byte_per_pixel * (width * y + x) + 1) = Hybrid_Median(valG);
            *(image_t_arr + byte_per_pixel * (width * y + x) + 0) = Hybrid_Median(valB);
        }
    }

    return 0;
}

int write_bmp(const char *fname_t)
{
    unsigned int file_size; // file size

    fp_t = fopen(fname_t, "wb");
    if (fp_t == NULL)
    {
        printf("fopen fname_t error\n");
        return -1;
    }

    // file size
    file_size = width * height * byte_per_pixel + rgb_raw_data_offset;
    header[2] = (unsigned char)(file_size & 0x000000ff);
    header[3] = (file_size >> 8)  & 0x000000ff;
    header[4] = (file_size >> 16) & 0x000000ff;
    header[5] = (file_size >> 24) & 0x000000ff;

    // width
    header[18] = width & 0x000000ff;
    header[19] = (width >> 8)  & 0x000000ff;
    header[20] = (width >> 16) & 0x000000ff;
    header[21] = (width >> 24) & 0x000000ff;

    // height
    header[22] = height & 0x000000ff;
    header[23] = (height >> 8)  & 0x000000ff;
    header[24] = (height >> 16) & 0x000000ff;
    header[25] = (height >> 24) & 0x000000ff;

    // bit per pixel
    header[28] = bit_per_pixel;

    // write header
    fwrite(header, sizeof(unsigned char), rgb_raw_data_offset, fp_t);

    // write image
    fwrite(image_t_arr, sizeof(unsigned char), (size_t)(long)width * height * byte_per_pixel, fp_t);

    fclose(fp_s);
    fclose(fp_t);

    return 0;
}



int main(int argc, char **argv)
{
    read_bmp("salt_noise.bmp"); // 24 bit gray level image
    sobel();
    write_bmp("salt_noise_hybrid.bmp");

    return 0;
}
