# package name : Verilator

# Verilator_FOUND
# Verilator_ROOT_DIR
# Verilator_INCLUDE_DIRS
# Verilator_SHARED_CPPS
# Verilator_DEFINITIONS
# Verilator_CFLAGS

set(possible_systemc_dir_name verilator)

if(NOT DEFINED Verilator_FIND_QUIETLY)
	set(Verilator_FIND_QUIETLY false)
endif()

if(NOT COMMAND append_list_each)
	message(FATAL_ERROR "please include utils.cmake before find_package()")
endif(NOT COMMAND append_list_each)

# find the Verilator library
foreach(dir_name ${possible_systemc_dir_name})
	append_list_each(systemc_path_list CMAKE_PREFIX_PATH /${dir_name})
	list(APPEND POSSIBLE_Verilator_PATHS ${systemc_path_list})
endforeach(dir_name)


foreach(SEARCH_PATH ${POSSIBLE_Verilator_PATHS})
	find_file(Verilator_INCLUDER_EXE verilator_bin
			PATH_SUFFIXES "bin"
			PATHS ${SEARCH_PATH} NO_DEFAULT_PATH)

	if(${Verilator_INCLUDER_EXE} STREQUAL "Verilator_INCLUDER_EXE-NOTFOUND")
		set(Verilator_FOUND false)
	else()
		set(Verilator_FOUND true)
	endif()

	if(${Verilator_FOUND})
			break()
	endif()
endforeach(SEARCH_PATH)


set(Verilator_FOUND ${Verilator_FOUND} CACHE BOOL "is found Verilator")

if(${Verilator_FOUND})

  get_filename_component(Verilator_BIN_DIR ${Verilator_INCLUDER_EXE} PATH)
  get_filename_component(Verilator_ROOT_DIR ${Verilator_BIN_DIR} PATH CACHE)

  if(NOT ${Verilator_FIND_QUIETLY})
	  message(STATUS "Verilator found : ${Verilator_ROOT_DIR}")
  endif(NOT ${Verilator_FIND_QUIETLY})

  # the SYSTEMC_VERSION is important that it defines the used interface
  set(Verilator_DEFINITIONS -DVL_PRINTF=printf -DVM_TRACE=1 -DVM_COVERAGE=0 -DTOP_TYPE=VGCD -DVL_USER_FINISH CACHE STRING "verilator definitions")
  set(Verilator_CFLAGS -MMD -Wno-char-subscripts -Wno-sign-compare -Wno-uninitialized -Wno-unused-but-set-variable -Wno-unused-parameter -Wno-unused-variable CACHE STRING "verilator cflags")

  set(VERILATOR_SHARE_DIR ${Verilator_ROOT_DIR}/share/verilator/include)
  set(VERILATOR_SHARE_DIR2 ${VERILATOR_SHARE_DIR}/vltstd)

  set(Verilator_INCLUDE_DIRS "${VERILATOR_SHARE_DIR}"
                             "${VERILATOR_SHARE_DIR2}"
					CACHE PATH "the included directory")

  set(Verilator_SHARED_CPPS ${VERILATOR_SHARE_DIR}/verilated_vcd_c.cpp ${VERILATOR_SHARE_DIR}/verilated.cpp CACHE FILE "the shared verilator c++ sources")

else(${Verilator_FOUND})
	  if(${Verilator_FIND_REQUIRED})
		  message(FATAL_ERROR "finds no Verilator")
	  endif(${Verilator_FIND_REQUIRED})
endif(${Verilator_FOUND})

